#
# Desktop.py -- Generic widgets Desktop GUI layout
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time
import math

from ginga.misc import Bunch, Callback
from ginga.gw import Widgets, Viewers


class Desktop(Callback.Callbacks):

    def __init__(self, app):
        super(Desktop, self).__init__()

        self.app = app
        self.logger = app.logger
        # for tabs
        self.tab = Bunch.caselessDict()
        self.tabcount = 0
        self.workspace = Bunch.caselessDict()

        self.toplevels = []
        self.node = {}
        self.node_idx = 0
        self._cur_dialogs = []

        for name in ('page-switch', 'page-close', 'all-closed'):
            self.enable_callback(name)

    # --- Workspace Handling ---

    def make_ws(self, name, group=1, show_tabs=True, show_border=False,
                detachable=False, tabpos=None, scrollable=True,
                closeable=False, wstype='tabs', use_toolbar=False):

        if (wstype in ('nb', 'ws', 'tabs')) and (not show_tabs):
            wstype = 'stack'

        ws = Workspace(name=name, wstype=wstype,
                       group=group, detachable=detachable,
                       use_toolbar=use_toolbar)
        ws.add_callback('page-switch', self.switch_page_cb)
        ws.add_callback('page-detach', self.page_detach_cb)
        ws.add_callback('page-close', self.page_close_cb)

        nb = ws.nb

        if tabpos is None:
            tabpos = 'top'

        if (wstype == 'tabs') and show_tabs:
            nb.set_tab_position(tabpos)

        self.workspace[name] = ws
        return ws

    def delete_ws(self, name):
        if not self.has_ws(name):
            raise ValueError("No workspace named '%s'" % (name))

        # NOTE: currently the Desktop object does not close the window
        # or tab--you must manage that yourself
        del self.workspace[name]

    def has_ws(self, name):
        return name in self.workspace

    def get_ws(self, name):
        return self.workspace[name]

    def get_nb(self, name):
        return self.workspace[name].nb

    def get_size(self, widget):
        return widget.get_size()

    def get_ws_size(self, name):
        w = self.get_nb(name)
        return self.get_size(w)

    def get_wsnames(self, group=1):
        res = []
        for name in self.workspace.keys():
            ws = self.workspace[name]
            if group is None:
                res.append(name)
            elif group == ws.group:
                res.append(name)
        return res

    def get_tabnames(self, group=1):
        res = []
        for name in self.tab.keys():
            bnch = self.tab[name]
            if group is None:
                res.append(name)
            elif group == bnch.group:
                res.append(name)
        return res

    def add_tab(self, wsname, widget, group, labelname, tabname=None,
                data=None):
        ws = self.get_ws(wsname)
        self.tabcount += 1
        if not tabname:
            tabname = labelname
            if tabname in self.tab:
                tabname = 'tab%d' % self.tabcount

        ws.add_tab(widget, title=labelname)
        bnch = Bunch.Bunch(widget=widget, name=labelname,
                           tabname=tabname, data=data,
                           group=group, wsname=wsname)
        self.tab[tabname] = bnch
        return bnch

    def _find_nb(self, tabname):
        widget = self.tab[tabname].widget
        for ws in self.workspace.values():
            nb = ws.nb
            page_num = nb.index_of(widget)
            if page_num < 0:
                continue
            return (nb, page_num)
        return (None, None)

    def _find_tab(self, widget):
        for key, bnch in self.tab.items():
            if widget == bnch.widget:
                return bnch
        return None

    def raise_tab(self, tabname):
        # construct a list of the tabs to raise in the order they
        # should be raised
        l = []
        name = tabname.lower()
        while name in self.tab:
            bnch = self.tab[name]
            l.insert(0, name)
            name = bnch.wsname.lower()
            if name in l:
                break

        # now raise those tabs
        for name in l:
            nb, index = self._find_nb(name)
            if (nb is not None) and (index >= 0):
                nb.set_index(index)

    def remove_tab(self, tabname):
        nb, index = self._find_nb(tabname)
        widget = self.tab[tabname].widget
        if (nb is not None) and (index >= 0):
            del self.tab[tabname]
            nb.remove(widget)

    def highlight_tab(self, tabname, onoff):
        nb, index = self._find_nb(tabname)
        if (nb is not None) and hasattr(nb, 'highlight_tab'):
            nb.highlight_tab(index, onoff)

    def show_dialog(self, dialog):
        dialog.show()
        # save a handle so widgets aren't garbage collected
        if dialog not in self._cur_dialogs:
            self._cur_dialogs.append(dialog)

    def hide_dialog(self, dialog):
        dialog.hide()

    def remove_dialog(self, dialog):
        dialog.hide()
        if dialog in self._cur_dialogs:
            self._cur_dialogs.remove(dialog)
        dialog.delete()

    def add_toplevel(self, bnch, wsname, width=700, height=700):
        topw = self.app.make_window(title=wsname)
        topw.resize(width, height)
        self.toplevels.append(topw)
        # TODO: don't ignore close, but handle workspace deletion
        topw.add_callback('close', lambda w: None)

        vbox = Widgets.VBox()
        vbox.set_border_width(0)
        topw.set_widget(vbox)

        vbox.add_widget(bnch.widget, stretch=1)
        topw.show()
        return topw

    def remove_toplevel(self, widget):
        widget.hide()
        if widget in self.toplevels:
            self.toplevels.remove(widget)
        widget.delete()

    def create_toplevel_ws(self, wsname, width, height,
                           group=2, x=None, y=None):
        # create main frame
        root = self.app.make_window()

        # TODO: this needs to be more sophisticated
        # root.set_title(wsname)
        ws = self.make_ws(wsname, wstype='tabs', use_toolbar=True)

        vbox = Widgets.VBox()
        vbox.set_border_width(0)

        vbox.add_widget(ws.widget)
        root.set_widget(vbox)

        root.resize(width, height)
        root.show()
        self.toplevels.append(root)

        if x is not None:
            root.move(x, y)

        return ws

    def _mk_take_tab_cb(self, tabname, to_nb):
        def _foo():
            nb, index = self._find_nb(tabname)
            widget = self.tab[tabname].widget
            if (nb is not None) and (index >= 0):
                nb.remove_tab(widget)
                to_nb.add_tab(widget, title=tabname)

        return _foo

    def _close_page(self, ws):
        num_children = ws.widget.num_children()
        if num_children == 0:
            del self.workspace[ws.name]
            root = ws.root
            ws.root = None
            root.delete()
        return True

    def page_detach_cb(self, ws, child):
        try:
            width, height = child.get_size()

            wsname = str(time.time())
            ws = self.create_toplevel_ws(wsname, width, height + 100)

            # get title of child
            try:
                title = child.extdata['tab_title']
            except KeyError:
                title = 'No Title'

            ws.add_tab(child, title=title)
        except Exception as e:
            print(str(e))

    def switch_page_cb(self, ws, child):
        self.logger.debug("page switch: %s" % str(child))
        bnch = self._find_tab(child)
        if bnch is not None:
            self.make_callback('page-switch', bnch.name, bnch.data)
        return False

    def page_close_cb(self, ws, child):
        self.logger.debug("page close: %s" % str(child))
        bnch = self._find_tab(child)
        if bnch is not None:
            self.make_callback('page-close', bnch.name, bnch.data)
        return False

    def make_desktop(self, layout, widget_dict=None):
        if widget_dict is None:
            widget_dict = {}
        self.node = {}
        self.node_idx = 0

        def process_common_params(kind, widget, params):
            subst_params = Bunch.Bunch(name=None, height=-1, width=-1,
                                       xpos=-1, ypos=-1, spacing=None)
            subst_params.update(params)
            params.update(subst_params)

            if params.name is not None:
                name = params.name
                widget_dict[params.name] = widget
            else:
                name = '%s_%d' % (kind, self.node_idx)
                self.node_idx += 1

            bnch = Bunch.Bunch(kind=kind, widget=widget,
                               name=name, params=params)
            self.node[name] = bnch

            # User is specifying the size of the widget
            if isinstance(widget, Widgets.WidgetBase):

                if params.spacing is not None:
                    widget.set_spacing(params.spacing)

                # directive to size widget
                if (params.width >= 0) or (params.height >= 0):
                    if params.width < 0:
                        width = widget.get_size()[0]
                    else:
                        width = params.width
                    if params.height < 0:
                        height = widget.get_size()[1]
                    else:
                        height = params.height
                    widget.resize(width, height)

                # User wants to place window somewhere
                if kind in ['top', 'dialog']:
                    if params.xpos >= 0:
                        widget.move(params.xpos, params.ypos)

            return bnch

        def make_widget(kind, params, args, pack):
            kind = kind.lower()

            # Process workspace parameters
            param_dict = Bunch.Bunch(name=None, title=None, height=-1,
                                     width=-1, group=1, show_tabs=True,
                                     show_border=False, scrollable=True,
                                     detachable=False, wstype='tabs',
                                     tabpos='top', use_toolbar=False)
            param_dict.update(params)
            params.update(param_dict)

            if kind == 'widget':
                widget = args[0]

            elif kind == 'ws':
                group = int(params.group)
                ws = self.make_ws(params.name, group=group,
                                  show_tabs=params.show_tabs,
                                  show_border=params.show_border,
                                  detachable=params.detachable,
                                  tabpos=params.tabpos,
                                  wstype=params.wstype,
                                  scrollable=params.scrollable,
                                  use_toolbar=params.use_toolbar)
                widget = ws.widget
                # debug(widget)

            # If a title was passed as a parameter, then make a frame to
            # wrap the widget using the title.
            if params.title:
                fr = Widgets.Frame(title=params.title)
                fr.set_widget(widget)
                pack(fr)
            else:
                pack(widget)

            res = []
            if (kind in ('ws', 'mdi', 'grid', 'stack')) and (len(args) > 0):
                # <-- Workspace specified a sub-layout.  We expect a list
                # of tabname, layout pairs--iterate over these and add them
                # to the workspace as tabs.
                for tabname, layout in args[0]:
                    def pack(w):
                        # ?why should group be the same as parent group?
                        self.add_tab(params.name, w, group,
                                     tabname, tabname.lower())

                    res.append((tabname, make(layout, pack)))

            bnch = process_common_params(kind, widget, params)
            if kind == 'ws':
                bnch.ws = ws

            # return [kind, params] + res
            return [kind, params, res]

        # Horizontal adjustable panel
        def hpanel(params, cols, pack):
            res = []
            if len(cols) >= 2:
                widget = Widgets.Splitter(orientation='horizontal')
                process_common_params('hpanel', widget, params)

                sizes = []
                for col in cols:
                    c = make(col, lambda w: widget.add_widget(w))
                    res.append(c)

                    # collect widths to set width of panes
                    s_params = col[1]
                    if 'width' in s_params:
                        sizes.append(s_params['width'])

                sizes = params.get('sizes', sizes)
                if len(sizes) == len(cols):
                    widget.set_sizes(sizes)
                    params.update(dict(sizes=sizes))

            elif len(cols) == 1:
                widget = Widgets.HBox()
                widget.set_border_width(0)
                process_common_params('hbox', widget, params)

                c = make(cols[0], lambda w: widget.add_widget(w, stretch=1))
                res.append(c)
                # widget.show()

            pack(widget)
            return ['hpanel', params] + res

        # Vertical adjustable panel
        def vpanel(params, rows, pack):
            res = []
            if len(rows) >= 2:
                widget = Widgets.Splitter(orientation='vertical')
                process_common_params('vpanel', widget, params)

                sizes = []
                for row in rows:
                    r = make(row, lambda w: widget.add_widget(w))
                    res.append(r)

                    # collect heights to set height of panes
                    s_params = row[1]
                    if 'height' in s_params:
                        sizes.append(s_params['height'])

                sizes = params.get('sizes', sizes)
                if len(sizes) == len(rows):
                    widget.set_sizes(sizes)
                    params.update(dict(sizes=sizes))

            elif len(rows) == 1:
                widget = Widgets.VBox()
                widget.set_border_width(0)
                process_common_params('vbox', widget, params)

                r = make(rows[0], lambda w: widget.add_widget(w, stretch=1))
                res.append(r)
                # widget.show()

            pack(widget)
            return ['vpanel', params] + res

        # Horizontal fixed array
        def hbox(params, cols, pack):
            widget = Widgets.HBox()
            widget.set_border_width(0)
            widget.set_spacing(0)

            res = []
            for dct in cols:
                if isinstance(dct, dict):
                    stretch = dct.get('stretch', 0)
                    col = dct.get('col', None)
                else:
                    # assume a list defining the col
                    stretch = 0
                    col = dct
                if col is not None:
                    c = make(col, lambda w: widget.add_widget(w,
                                                              stretch=stretch))
                    c = dict(col=c, stretch=stretch)
                    res.append(c)

            process_common_params('hbox', widget, params)
            pack(widget)
            return ['hbox', params] + res

        # Vertical fixed array
        def vbox(params, rows, pack):
            widget = Widgets.VBox()
            widget.set_border_width(0)
            widget.set_spacing(0)

            res = []
            for dct in rows:
                if isinstance(dct, dict):
                    stretch = dct.get('stretch', 0)
                    row = dct.get('row', None)
                else:
                    # assume a list defining the row
                    stretch = 0
                    row = dct
                if row is not None:
                    r = make(row, lambda w: widget.add_widget(w,
                                                              stretch=stretch))
                    r = dict(row=r, stretch=stretch)
                    res.append(r)

            process_common_params('vbox', widget, params)
            pack(widget)
            return ['vbox', params] + res

        # top-level dialog box
        def dialog(params, rows, pack):
            if len(self.toplevels) == 0:
                raise ValueError("Need a top-level window to parent a dialog")

            # any old top-level will do, for now...
            top_w = self.toplevels[0]
            widget = Widgets.Dialog(parent=top_w)
            container = widget.get_content_area()

            res = []
            for dct in rows:
                if isinstance(dct, dict):
                    stretch = dct.get('stretch', 1)
                    row = dct.get('row', None)
                else:
                    # assume a list defining the row
                    stretch = 1
                    row = dct
                if row is not None:
                    r = make(row,
                             lambda w: container.add_widget(w,
                                                            stretch=stretch))
                    r = dict(row=r, stretch=stretch)
                    res.append(r)

            process_common_params('dialog', widget, params)

            widget.show()
            return ['dialog', params] + res

        # top-level window
        def toplevel(params, rows, pack):
            widget = self.app.make_window()
            self.toplevels.append(widget)

            if len(rows) != 1:
                raise ValueError("top-level windows can only have 1 child")

            res = []
            dct = rows[0]
            if isinstance(dct, dict):
                row = dct.get('row', None)
            else:
                # assume a list defining the row
                row = dct
            if row is not None:
                r = make(row, lambda w: widget.set_widget(w))
                r = dict(row=r)
                res.append(r)

            process_common_params('top', widget, params)

            widget.show()
            return ['top', params] + res

        # Sequence of separate top-level items
        def seq(params, cols, pack):
            def mypack(w):
                w_top = self.app.make_window()
                # Ask the size of the widget that wants to get packed
                # and resize the top-level to fit
                wd, ht = w.get_size()
                w_top.resize(wd, ht)
                w_top.set_widget(w)
                self.toplevels.append(w_top)
                w_top.show()
                process_common_params('top', w_top, params)

            res = []
            for dct in cols:
                if isinstance(dct, dict):
                    col = dct.get('col', None)
                else:
                    # assume a list defining the col
                    col = dct
                if col is not None:
                    res.append(make(col, mypack))

            # presumably, the passed-in pack is a no-op, so this
            # should do nothing
            widget = Widgets.Label("Placeholder")
            pack(widget)
            return ['seq', params] + res

        def make(constituents, pack):
            kind = constituents[0]
            params = Bunch.Bunch(constituents[1])
            if len(constituents) > 2:
                rest = constituents[2:]
            else:
                rest = []

            if kind == 'vpanel':
                res = vpanel(params, rest, pack)
            elif kind == 'hpanel':
                res = hpanel(params, rest, pack)
            elif kind == 'vbox':
                res = vbox(params, rest, pack)
            elif kind == 'hbox':
                res = hbox(params, rest, pack)
            elif kind == 'dialog':
                res = dialog(params, rest, pack)
            elif kind == 'top':
                res = toplevel(params, rest, pack)
            elif kind == 'seq':
                res = seq(params, rest, pack)
            elif kind in ('ws', 'mdi', 'widget'):
                res = make_widget(kind, params, rest, pack)

            return res

        self.layout = make(layout, lambda w: None)
        return self.layout

    def record_sizes(self):
        """Record sizes of all container widgets in the layout.
        The sizes are recorded in the `params` mappings in the layout.
        """
        for rec in self.node.values():
            w = rec.widget
            wd, ht = w.get_size()

            rec.params.update(dict(name=rec.name, width=wd, height=ht))

            if rec.kind in ('hpanel', 'vpanel'):
                sizes = w.get_sizes()
                rec.params.update(dict(sizes=sizes))

            if rec.kind in ('ws',):
                # record ws type
                # TODO: record positions of subwindows
                rec.params.update(dict(wstype=rec.ws.wstype))

            if rec.kind == 'top':
                # record position for top-level widgets as well
                x, y = w.get_pos()
                if x is not None and y is not None:
                    # we don't always get reliable information about position
                    rec.params.update(dict(xpos=x, ypos=y))

    def write_layout_conf(self, lo_file, layout=None):
        if layout is None:
            layout = self.layout

        import pprint
        self.record_sizes()

        # write layout
        with open(lo_file, 'w') as out_f:
            pprint.pprint(layout, out_f)

    def read_layout_conf(self, lo_file):
        import ast

        # read layout
        with open(lo_file, 'r') as in_f:
            buf = in_f.read()

        layout = ast.literal_eval(buf)
        return layout

    def build_desktop(self, layout, lo_file=None, widget_dict=None):
        if lo_file is not None:
            alt_layout = layout
            try:
                layout = self.read_layout_conf(lo_file)

            except Exception as e:
                self.logger.info("Error reading saved layout: %s" % (str(e)))
                layout = alt_layout

        return self.make_desktop(layout, widget_dict=widget_dict)


# WORKSPACES #####

class Workspace(Widgets.WidgetBase):

    def __init__(self, name, wstype='tab', group=0, detachable=False,
                 use_toolbar=False):
        super(Workspace, self).__init__()

        self.name = name
        self.wstype = wstype
        self.wstypes = ['tabs', 'mdi', 'stack', 'grid']
        self.vbox = Widgets.VBox()
        self.vbox.set_spacing(0)
        self.vbox.set_border_width(0)
        # for now
        self.widget = self.vbox
        self.nb = None
        self.group = group
        self.detachable = detachable
        self.mdi_menu = None
        self.wstypes_c = ['Tabs', 'Grid', 'MDI', 'Stack']
        self.wstypes_l = ['tabs', 'grid', 'mdi', 'stack']
        self.extdata = Bunch.Bunch(toolbar=None)
        self.toolbar = None

        if use_toolbar:
            toolbar = Widgets.Toolbar(orientation='horizontal')
            toolbar.set_border_width(0)
            self.vbox.add_widget(toolbar, stretch=0)
            self.toolbar = toolbar
            self.extdata.toolbar = toolbar

            ws_menu = toolbar.add_menu("Workspace")
            item = ws_menu.add_name("Close")
            item.add_callback('activated', self._close_menuitem_cb)

        self._set_wstype(wstype)
        self.vbox.add_widget(self.nb, stretch=1)

        if use_toolbar:
            cbox = Widgets.ComboBox()
            for _wstype in self.wstypes_c:
                cbox.append_text(_wstype)
            idx = self.wstypes_l.index(wstype)
            cbox.set_index(idx)
            cbox.add_callback('activated', self.config_wstype_cb)
            cbox.set_tooltip("Set workspace type")
            self.toolbar.add_widget(cbox)

            mdi_menu = self.toolbar.add_menu("MDI")

            item = mdi_menu.add_name("Tile Panes")
            item.add_callback('activated',
                              lambda *args: self.tile_panes_cb())

            item = mdi_menu.add_name("Cascade Panes")
            item.add_callback('activated',
                              lambda *args: self.cascade_panes_cb())
            self.mdi_menu = mdi_menu
            self.extdata.mdi_menu = mdi_menu
            self._update_mdi_menu()

        for name in ('page-switch', 'page-detach', 'page-close',
                     'page-removed', 'page-added', 'ws-close'):
            self.enable_callback(name)

    def _set_wstype(self, wstype):
        if wstype in ('tabs', 'nb', 'ws'):
            wstype = 'tabs'
            self.nb = Widgets.TabWidget(detachable=self.detachable,
                                        reorderable=True, group=self.group)

        elif wstype == 'mdi':
            self.nb = Widgets.MDIWidget(mode='mdi')

        elif wstype == 'stack':
            self.nb = Widgets.StackWidget()

        elif wstype == 'grid':
            self.nb = SymmetricGridWidget()

        self._update_mdi_menu()
        if self.nb.has_callback('page-switch'):
            self.nb.add_callback('page-switch', self._switch_page_cb)
        if self.nb.has_callback('page-detach'):
            self.nb.add_callback('page-detach', self._detach_page_cb)
        if self.nb.has_callback('page-close'):
            self.nb.add_callback('page-close', self._close_page_cb)
        if self.nb.has_callback('widget-added'):
            self.nb.add_callback('widget-added', self._page_added_cb)
        if self.nb.has_callback('widget-removed'):
            self.nb.add_callback('widget-removed', self._page_removed_cb)

        self.wstype = wstype

    def _update_mdi_menu(self):
        if self.mdi_menu is None:
            return
        if isinstance(self.nb, Widgets.MDIWidget) and self.nb.true_mdi:
            self.mdi_menu.set_enabled(True)
        else:
            self.mdi_menu.set_enabled(False)

    def tile_panes_cb(self):
        if isinstance(self.nb, Widgets.MDIWidget) and self.nb.true_mdi:
            self.nb.tile_panes()

    def cascade_panes_cb(self):
        if isinstance(self.nb, Widgets.MDIWidget) and self.nb.true_mdi:
            self.nb.cascade_panes()

    def tabstoggle_cb(self, tf):
        if isinstance(self.nb, Widgets.MDIWidget) and self.nb.true_mdi:
            self.nb.use_tabs(tf)

    def _switch_page_cb(self, nb, child):
        self.focus_index()
        # redirect widget callback into workspace callback
        self.make_callback('page-switch', child)

    def _detach_page_cb(self, nb, child):
        # redirect widget callback into workspace callback
        self.make_callback('page-detach', child)

    def _page_added_cb(self, nb, child):
        # redirect widget callback into workspace callback
        self.make_callback('page-added', child)

    def _page_removed_cb(self, nb, child):
        # redirect widget callback into workspace callback
        self.make_callback('page-removed', child)

    def _close_page_cb(self, nb, child):
        # redirect widget callback into workspace callback
        self.make_callback('page-close', child)

    def _close_menuitem_cb(self, *args):
        self.make_callback('ws-close')

    def num_pages(self):
        return self.nb.num_children()

    def configure_wstype(self, wstype):
        old_widget = self.nb
        self.vbox.remove(old_widget)

        # remember which tab was on top
        idx = old_widget.get_index()

        self._set_wstype(wstype)
        self.vbox.add_widget(self.nb, stretch=1)

        for child in list(old_widget.get_children()):
            # TODO: sort by previous index so they get added to the
            # new widget in the same order
            title = child.extdata.get('tab_title', '')
            old_widget.remove(child)
            self.nb.add_widget(child, title=title)
            child.show()

        # restore focus to widget that was on top
        if idx >= 0:
            self.nb.set_index(idx)
        self.focus_index()

    def config_wstype_cb(self, w, idx):
        wstype = self.wstypes_l[idx]
        self.configure_wstype(wstype)

    def cycle_wstype(self):
        idx = self.wstypes.index(self.wstype)
        idx = (idx + 1) % len(self.wstypes)
        wstype = self.wstypes[idx]
        self.configure_wstype(wstype)

    def focus_index(self):
        def _f(widget):
            # TODO: this probably just ought to check whether a widget
            # *can take* focus
            if isinstance(widget, Viewers.GingaViewerWidget):
                widget.focus()
                return True
            # widget can't take focus.  If it is a container widget,
            # check its children
            if isinstance(widget, Widgets.ContainerBase):
                for child in widget.get_children():
                    if _f(child):
                        return True
                return False
            return False

        cur_idx = self.nb.get_index()
        try:
            child = self.nb.index_to_widget(cur_idx)
        except IndexError:
            return
        _f(child)

    def to_next(self):
        num_tabs = self.nb.num_children()
        cur_idx = self.nb.get_index()
        new_idx = (cur_idx + 1) % num_tabs
        self.nb.set_index(new_idx)
        self.focus_index()

    def to_previous(self):
        num_tabs = self.nb.num_children()
        new_idx = self.nb.get_index() - 1
        if new_idx < 0:
            new_idx = max(num_tabs - 1, 0)
        self.nb.set_index(new_idx)
        self.focus_index()

    def add_tab(self, child, title=''):
        self.nb.add_widget(child, title=title)

    def remove_tab(self, child):
        self.nb.remove(child)


class SymmetricGridWidget(Widgets.GridBox):
    """Custom widget for grid-type workspace that has the API of a
    tab-like widget.
    """

    def __init__(self):
        super(SymmetricGridWidget, self).__init__()

        self.set_margins(0, 0, 0, 0)
        self.set_spacing(2)
        self.cur_index = 0

        for name in ['page-switch']:
            self.enable_callback(name)

    def _calc_dims(self, num_widgets):
        rows = int(round(math.sqrt(num_widgets)))
        cols = rows
        if rows ** 2 < num_widgets:
            cols += 1
        return rows, cols

    def _relayout(self, widgets):
        # remove all the old widgets
        # TEMP: this call causes a recursive loop to happen
        # self.remove_all()
        children = list(self.get_children())
        for child in children:
            super(SymmetricGridWidget, self).remove(child)

        # calculate number of rows and cols, try to maintain a square
        # TODO: take into account the window geometry
        num_widgets = len(widgets)
        rows, cols = self._calc_dims(num_widgets)

        self.resize_grid(rows, cols)

        # add them back in, in a grid
        for i in range(0, rows):
            for j in range(0, cols):
                index = i * cols + j
                if index < num_widgets:
                    child = widgets[index]
                    super(SymmetricGridWidget, self).add_widget(child, i, j,
                                                                stretch=1)

    def add_widget(self, child, title=''):
        num_widgets = self.num_children()
        o_rows, o_cols = self._calc_dims(num_widgets)

        num_widgets += 1
        rows, cols = self._calc_dims(num_widgets)

        # attach title to child
        child.extdata.tab_title = title

        if (o_rows, o_cols) == (rows, cols):
            # size of matrix has not changed--
            # we can be a little efficient here and skip rebuilding
            j = num_widgets - ((rows - 1) * cols) - 1
            super(SymmetricGridWidget, self).add_widget(child, rows - 1, j,
                                                        stretch=1)
        else:
            widgets = list(self.get_children())
            widgets.append(child)
            self._relayout(widgets)

        self.make_callback('widget-added', child)

    def remove(self, child, delete=False):

        widgets = list(self.get_children())
        try:
            idx = widgets.index(child)

        except IndexError:
            raise ValueError("child not in container")

        o_rows, o_cols = self._calc_dims(len(widgets))

        super(SymmetricGridWidget, self).remove(child)

        num_widgets = self.num_children()
        rows, cols = self._calc_dims(num_widgets)

        # if size of matrix has not changed AND user only deleted from
        # the end--we can be a little efficient here and skip rebuilding
        if (child != widgets[-1]) or ((o_rows, o_cols) != (rows, cols)):
            widgets = list(self.get_children())
            self._relayout(widgets)

        if idx < self.cur_index:
            # deleted a widget lower than the current index:
            # current index just shifts down
            self.cur_index -= 1

        elif idx == self.cur_index:
            # deleted same widget as current index
            self.cur_index = -1
            num_children = self.num_children()
            if 0 <= idx < num_children:
                # we can keep same index
                self.set_index(idx)
            else:
                # pick acceptable index
                idx = max(0, min(idx - 1, num_children - 1))
                if num_children > 0:
                    self.set_index(idx)

    def get_index(self):
        return self.cur_index

    def set_index(self, idx):
        old_index = self.cur_index
        num_children = self.num_children()
        if 0 <= idx < num_children:
            self.cur_index = idx
            child = self.children[idx]
            # child.focus()

            if old_index != idx:
                self.make_callback('page-switch', child)

        else:
            raise ValueError("index (%d) inconsistent with number of children (%d)" %
                             (idx, num_children))

    def index_of(self, child):
        children = self.get_children()
        try:
            return children.index(child)
        except (IndexError, ValueError) as e:
            return -1

    def index_to_widget(self, idx):
        children = self.get_children()
        return children[idx]

# END



#
# wcs.py -- calculations based on world coordinate system.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""This module handles calculations based on world coordinate system."""
import math
from collections import OrderedDict

import numpy as np

from ginga.misc import Bunch

__all__ = ['hmsToDeg', 'dmsToDeg', 'decTimeToDeg', 'degToHms', 'degToDms',
           'arcsecToDeg', 'hmsStrToDeg', 'dmsStrToDeg', 'raDegToString',
           'decDegToString', 'trans_coeff', 'eqToEq2000',
           'get_xy_rotation_and_scale', 'get_rotation_and_scale',
           'get_relative_orientation', 'simple_wcs', 'deg2fmt', 'dispos',
           'deltaStarsRaDecDeg1', 'deltaStarsRaDecDeg2', 'get_starsep_RaDecDeg',
           'add_offset_radec', 'get_RaDecOffsets', 'lon_to_deg', 'lat_to_deg',
           ]


class WCSError(Exception):
    pass


degPerHMSHour = 15.0        # 360/24
degPerHMSMin = 0.25         # 360.0/24/60
degPerHMSSec = 1.0 / 240.0  # 360.0/24/60/60

degPerDmsMin = 1.0 / 60.0
degPerDmsSec = 1.0 / 3600.0

HMSHourPerDeg = 1.0 / 15.0
HMSMinPerDeg = 4.0
HMSSecPerDeg = 240.0


def hmsToDeg(h, m, s):
    """Convert RA hours, minutes, seconds into an angle in degrees."""
    return h * degPerHMSHour + m * degPerHMSMin + s * degPerHMSSec


def dmsToDeg(sign, deg, min, sec):
    """Convert dec sign, degrees, minutes, seconds into a signed angle in
    degrees."""
    return sign * (deg + min * degPerDmsMin + sec * degPerDmsSec)


def decTimeToDeg(sign_sym, deg, min, sec):
    """Convert dec sign, degrees, minutes, seconds into a signed angle in
    degrees.

    ``sign_sym`` may represent negative as either '-' or numeric -1."""
    if sign_sym == -1 or sign_sym == '-':
        sign = -1
    else:
        sign = 1
    return dmsToDeg(sign, deg, min, sec)


def degToHms(ra):
    """Converts the ra (in degrees) to HMS three tuple.
    H and M are in integer and the S part is in float.
    """
    assert (ra >= 0.0), WCSError("RA (%f) is negative" % (ra))
    assert ra < 360.0, WCSError("RA (%f) > 360.0" % (ra))
    rah = ra / degPerHMSHour
    ramin = (ra % degPerHMSHour) * HMSMinPerDeg
    rasec = (ra % degPerHMSMin) * HMSSecPerDeg
    return (int(rah), int(ramin), rasec)


def degToDms(dec, isLatitude=True):
    """Convert the dec, in degrees, to an (sign,D,M,S) tuple.
    D and M are integer, and sign and S are float.
    """
    if isLatitude:
        assert dec <= 90, WCSError("DEC (%f) > 90.0" % (dec))
        assert dec >= -90, WCSError("DEC (%f) < -90.0" % (dec))

    if dec < 0.0:
        sign = -1.0
    else:
        sign = 1.0
    dec = dec * sign

    #mnt = (dec % 1.0) * 60.0
    #sec = (dec % (1.0/60.0)) * 3600.0
    # this calculation with return values produces conversion problem.
    # e.g. dec +311600.00 -> 31.2666666667 degree
    # deg=31 min=15 sec=60 instead deg=31 min=16 sec=0.0
    # bug fixed
    mnt, sec = divmod(dec * 3600, 60)
    deg, mnt = divmod(mnt, 60)

    return (int(sign), int(deg), int(mnt), sec)


def arcsecToDeg(arcsec):
    """Convert numeric arcseconds (aka DMS seconds) to degrees of arc.
    """
    return arcsec * degPerDmsSec


def hmsStrToDeg(ra):
    """Convert a string representation of RA into a float in degrees."""
    hour, min, sec = ra.split(':')
    ra_deg = hmsToDeg(int(hour), int(min), float(sec))
    return ra_deg


def dmsStrToDeg(dec):
    """Convert a string representation of DEC into a float in degrees."""
    sign_deg, min, sec = dec.split(':')
    sign = sign_deg[0:1]
    if sign not in ('+', '-'):
        sign = '+'
        deg = sign_deg
    else:
        deg = sign_deg[1:]
    dec_deg = decTimeToDeg(sign, int(deg), int(min), float(sec))
    return dec_deg


def raDegToString(ra_deg, format='%02d:%02d:%06.3f'):
    if ra_deg > 360.0:
        ra_deg = math.fmod(ra_deg, 360.0)

    ra_hour, ra_min, ra_sec = degToHms(ra_deg)
    return format % (ra_hour, ra_min, ra_sec)


def decDegToString(dec_deg, format='%s%02d:%02d:%05.2f'):
    sign, dec_degree, dec_min, dec_sec = degToDms(dec_deg)
    if sign > 0:
        sign_sym = '+'
    else:
        sign_sym = '-'
    return format % (sign_sym, int(dec_degree), int(dec_min), dec_sec)


def trans_coeff(eq, x, y, z):
    """This function is provided by MOKA2 Development Team (1996.xx.xx)
    and used in SOSS system."""
    tt = (eq - 2000.0) / 100.0

    zeta = 2306.2181 * tt + 0.30188 * tt * tt + 0.017998 * tt * tt * tt
    zetto = 2306.2181 * tt + 1.09468 * tt * tt + 0.018203 * tt * tt * tt
    theta = 2004.3109 * tt - 0.42665 * tt * tt - 0.041833 * tt * tt * tt

    zeta = math.radians(zeta) / 3600.0
    zetto = math.radians(zetto) / 3600.0
    theta = math.radians(theta) / 3600.0

    p11 = (math.cos(zeta) * math.cos(theta) * math.cos(zetto) -
           math.sin(zeta) * math.sin(zetto))
    p12 = (-math.sin(zeta) * math.cos(theta) * math.cos(zetto) -
           math.cos(zeta) * math.sin(zetto))
    p13 = -math.sin(theta) * math.cos(zetto)
    p21 = (math.cos(zeta) * math.cos(theta) * math.sin(zetto) +
           math.sin(zeta) * math.cos(zetto))
    p22 = (-math.sin(zeta) * math.cos(theta) * math.sin(zetto) +
           math.cos(zeta) * math.cos(zetto))
    p23 = -math.sin(theta) * math.sin(zetto)
    p31 = math.cos(zeta) * math.sin(theta)
    p32 = -math.sin(zeta) * math.sin(theta)
    p33 = math.cos(theta)

    return (p11, p12, p13, p21, p22, p23, p31, p32, p33)


def eqToEq2000(ra_deg, dec_deg, eq):
    """Convert Eq to Eq 2000."""
    ra_rad = math.radians(ra_deg)
    dec_rad = math.radians(dec_deg)

    x = math.cos(dec_rad) * math.cos(ra_rad)
    y = math.cos(dec_rad) * math.sin(ra_rad)
    z = math.sin(dec_rad)

    p11, p12, p13, p21, p22, p23, p31, p32, p33 = trans_coeff(eq, x, y, z)

    x0 = p11 * x + p21 * y + p31 * z
    y0 = p12 * x + p22 * y + p32 * z
    z0 = p13 * x + p23 * y + p33 * z

    new_dec = math.asin(z0)
    if x0 == 0.0:
        new_ra = math.pi / 2.0
    else:
        new_ra = math.atan(y0 / x0)

    if ((y0 * math.cos(new_dec) > 0.0 and x0 * math.cos(new_dec) <= 0.0) or
            (y0 * math.cos(new_dec) <= 0.0 and x0 * math.cos(new_dec) < 0.0)):
        new_ra += math.pi

    elif new_ra < 0.0:
        new_ra += 2.0 * math.pi

    #new_ra = new_ra * 12.0 * 3600.0 / math.pi
    new_ra_deg = new_ra * 12.0 / math.pi * 15.0

    #new_dec = new_dec * 180.0 * 3600.0 / math.pi
    new_dec_deg = new_dec * 180.0 / math.pi

    return (new_ra_deg, new_dec_deg)


def get_xy_rotation_and_scale(header):
    """
    CREDIT: See IDL code at
    http://www.astro.washington.edu/docs/idl/cgi-bin/getpro/library32.html?GETROT
    """

    def calc_from_cd(cd1_1, cd1_2, cd2_1, cd2_2):

        # TODO: Check if first coordinate in CTYPE is latitude
        # if (ctype EQ 'DEC-') or (strmid(ctype, 1) EQ 'LAT')  then $
        #    cd = reverse(cd,1)

        det = cd1_1 * cd2_2 - cd1_2 * cd2_1
        if det < 0:
            sgn = -1
        else:
            sgn = 1
        ## if det > 0:
        ##     raise ValueError("Astrometry is for a right-handed coordinate system")

        if (cd2_1 == 0.0) or (cd1_2 == 0.0):
            # Unrotated coordinates?
            xrot = 0.0
            yrot = 0.0
            cdelt1 = cd1_1
            cdelt2 = cd2_2
        else:
            xrot = math.atan2(sgn * cd1_2, sgn * cd1_1)
            yrot = math.atan2(-cd2_1, cd2_2)

            cdelt1 = sgn * math.sqrt(cd1_1**2 + cd1_2**2)
            cdelt2 = math.sqrt(cd1_1**2 + cd2_1**2)

        return xrot, yrot, cdelt1, cdelt2

    def calc_from_crota():
        try:
            crota1 = float(header['CROTA1'])
            xrot = crota1
        except KeyError:
            xrot = None

        try:
            crota2 = float(header['CROTA2'])
            yrot = crota2
        except KeyError:
            yrot = 0.0

        if xrot is None:
            xrot = yrot

        cdelt1 = float(header.get('CDELT1', 1.0))
        cdelt2 = float(header.get('CDELT2', 1.0))

        return xrot, yrot, cdelt1, cdelt2

    # 1st, check for presence of PC matrix
    try:
        pc1_1 = header['PC1_1']
        pc1_2 = header['PC1_2']
        pc2_1 = header['PC2_1']
        pc2_2 = header['PC2_2']

        cdelt1 = float(header['CDELT1'])
        cdelt2 = float(header['CDELT2'])

        cd1_1, cd1_2 = pc1_1 * cdelt1, pc1_2 * cdelt1
        cd2_1, cd2_2 = pc2_1 * cdelt2, pc2_2 * cdelt2

        xrot, yrot, cdelt1p, cdelt2p = calc_from_cd(pc1_1, pc1_2,
                                                    pc2_1, pc2_2)

    except KeyError:
        # 2nd, check for presence of CD matrix
        try:
            cd1_1 = header['CD1_1']
            cd1_2 = header['CD1_2']
            cd2_1 = header['CD2_1']
            cd2_2 = header['CD2_2']
            xrot, yrot, cdelt1, cdelt2 = calc_from_cd(cd1_1, cd1_2,
                                                      cd2_1, cd2_2)

        except KeyError:
            # 3rd, check for presence of CROTA keyword
            #  (or default is north=up)
            xrot, yrot, cdelt1, cdelt2 = calc_from_crota()

    xrot, yrot = math.degrees(xrot), math.degrees(yrot)

    return ((xrot, yrot), (cdelt1, cdelt2))


def get_rotation_and_scale(header, skew_threshold=0.001):
    """Calculate rotation and CDELT."""
    ((xrot, yrot),
     (cdelt1, cdelt2)) = get_xy_rotation_and_scale(header)

    if math.fabs(xrot) - math.fabs(yrot) > skew_threshold:
        raise ValueError("Skew detected: xrot=%.4f yrot=%.4f" % (
            xrot, yrot))
    rot = yrot

    lonpole = float(header.get('LONPOLE', 180.0))
    if lonpole != 180.0:
        rot += 180.0 - lonpole

    return (rot, cdelt1, cdelt2)


def get_relative_orientation(image, ref_image):
    """Computes the relative orientation and scale of an image to a reference
    image.

    Parameters
    ----------
    image
        AstroImage based object

    ref_image
        AstroImage based object

    Returns
    -------
    result
        Bunch object containing the relative scale in x and y
        and the relative rotation in degrees.

    """
    # Get reference image rotation and scale
    header = ref_image.get_header()
    ((xrot_ref, yrot_ref),
     (cdelt1_ref, cdelt2_ref)) = get_xy_rotation_and_scale(header)

    scale_x, scale_y = math.fabs(cdelt1_ref), math.fabs(cdelt2_ref)

    # Get rotation and scale of image
    header = image.get_header()
    ((xrot, yrot),
     (cdelt1, cdelt2)) = get_xy_rotation_and_scale(header)

    # Determine relative scale of this image to the reference
    rscale_x = math.fabs(cdelt1) / scale_x
    rscale_y = math.fabs(cdelt2) / scale_y

    # Figure out rotation relative to our orientation
    rrot_dx, rrot_dy = xrot - xrot_ref, yrot - yrot_ref

    # flip_x = False
    # flip_y = False

    # ## # Flip X due to negative CDELT1
    # ## if np.sign(cdelt1) < 0:
    # ##     flip_x = True

    # ## # Flip Y due to negative CDELT2
    # ## if np.sign(cdelt2) < 0:
    # ##     flip_y = True

    # # Optomization for 180 rotations
    # if np.isclose(math.fabs(rrot_dx), 180.0):
    #     flip_x = not flip_x
    #     rrot_dx = 0.0
    # if np.isclose(math.fabs(rrot_dy), 180.0):
    #     flip_y = not flip_y
    #     rrot_dy = 0.0

    rrot_deg = max(rrot_dx, rrot_dy)

    res = Bunch.Bunch(rscale_x=rscale_x, rscale_y=rscale_y,
                      rrot_deg=rrot_deg)
    return res


def simple_wcs(px_x, px_y, ra_deg, dec_deg, px_scale_deg_px, rot_deg,
               cdbase=[1, 1]):
    """Calculate a set of WCS keywords for a 2D simple instrument FITS
    file with a 'standard' RA/DEC pixel projection.

    Parameters
    ----------
    px_x
        (ZERO-based) reference pixel of field in X
        (usually center of field)

    px_y
        (ZERO-based) reference pixel of field in Y
        (usually center of field)

    ra_deg
        RA (in deg) for the reference point

    dec_deg
        DEC (in deg) for the reference point

    px_scale_deg_px
        Pixel scale (deg/pixel); can be a tuple for different x,y scales

    rot_deg
        Rotation angle of the field (in deg)

    cdbase
        CD base

    Returns
    -------
    res : dict
        Ordered dictionary object containing WCS headers.

    """
    # center of the projection
    crpix = (px_x + 1, px_y + 1)  # pixel position (WCS is 1 based)
    crval = (ra_deg, dec_deg)  # RA, Dec (degrees)

    # image scale in deg/pix
    cdelt = np.array(cdbase) * np.array(px_scale_deg_px)

    # Create rotation matrix for position angle of north (radians E of N)
    rot_rad = np.radians(rot_deg)
    cpa = np.cos(rot_rad)
    spa = np.sin(rot_rad)
    # a) clockwise rotation
    pc = np.array([[cpa, spa], [-spa, cpa]])
    # b) counter clockwise
    #pc = np.array([[cpa, -spa], [spa, cpa]])

    #cd = pc * cdelt

    res = OrderedDict((('CRVAL1', crval[0]),
                       ('CRVAL2', crval[1]),
                       ('CRPIX1', crpix[0]),
                       ('CRPIX2', crpix[1]),
                       ('CUNIT1', 'deg'),
                       ('CUNIT2', 'deg'),
                       ('CTYPE1', 'RA---TAN'),
                       ('CTYPE2', 'DEC--TAN'),
                       ('RADESYS', 'FK5'),
                       # Either PC + CDELT or CD should appear
                       # PC + CDELT seems to be the preferred approach
                       # according to the Calabretta papers
                       ('CDELT1', cdelt[0]),
                       ('CDELT2', cdelt[1]),
                       ('PC1_1', pc[0, 0]),
                       ('PC1_2', pc[0, 1]),
                       ('PC2_1', pc[1, 0]),
                       ('PC2_2', pc[1, 1])
                       ## ('CD1_1' , cd[0, 0]),
                       ## ('CD1_2' , cd[0, 1]),
                       ## ('CD2_1' , cd[1, 0]),
                       ## ('CD2_2' , cd[1, 1]),
                       ))
    return res


def deg2fmt(ra_deg, dec_deg, format):
    """Format coordinates."""

    rhr, rmn, rsec = degToHms(ra_deg)
    dsgn, ddeg, dmn, dsec = degToDms(dec_deg)

    if format == 'hms':
        return rhr, rmn, rsec, dsgn, ddeg, dmn, dsec

    elif format == 'str':
        #ra_txt = '%02d:%02d:%06.3f' % (rhr, rmn, rsec)
        ra_txt = '%d:%02d:%06.3f' % (rhr, rmn, rsec)
        if dsgn < 0:
            dsgn = '-'
        else:
            dsgn = '+'
        #dec_txt = '%s%02d:%02d:%05.2f' % (dsgn, ddeg, dmn, dsec)
        dec_txt = '%s%d:%02d:%05.2f' % (dsgn, ddeg, dmn, dsec)
        return ra_txt, dec_txt


def dispos(dra0, decd0, dra, decd):
    """Compute distance and position angle solving a spherical
    triangle (no approximations).

    Source/credit: Skycat
    Author: A.P. Martinez

    Parameters
    ----------
    dra0 : float
        Center RA in decimal degrees.

    decd0 : float
        Center DEC in decimal degrees.

    dra : float
        Point RA in decimal degrees.

    decd : float
        Point DEC in decimal degrees.

    Returns
    -------
    phi : float
        Phi in degrees (East of North).

    dist : float
        Distance in arcmin.

    """
    radian = 180.0 / math.pi

    # coord transformed in radians
    alf = dra / radian
    alf0 = dra0 / radian
    del_ = decd / radian
    del0 = decd0 / radian

    sd0 = math.sin(del0)
    sd = math.sin(del_)
    cd0 = math.cos(del0)
    cd = math.cos(del_)
    cosda = math.cos(alf - alf0)
    cosd = sd0 * sd + cd0 * cd * cosda
    dist = math.acos(cosd)
    phi = 0.0
    if dist > 0.0000004:
        sind = math.sin(dist)
        cospa = (sd * cd0 - cd * sd0 * cosda) / sind
        #if cospa > 1.0:
        #    cospa=1.0
        if math.fabs(cospa) > 1.0:
            # 2005-06-02: fix from awicenec@eso.org
            cospa = cospa / math.fabs(cospa)
        sinpa = cd * math.sin(alf - alf0) / sind
        phi = math.acos(cospa) * radian
        if sinpa < 0.0:
            phi = 360.0 - phi
    dist *= radian
    dist *= 60.0
    if decd0 == 90.0:
        phi = 180.0
    if decd0 == -90.0:
        phi = 0.0
    return (phi, dist)


def deltaStarsRaDecDeg1(ra1_deg, dec1_deg, ra2_deg, dec2_deg):
    """Spherical triangulation."""
    phi, dist = dispos(ra1_deg, dec1_deg, ra2_deg, dec2_deg)
    return arcsecToDeg(dist * 60.0)


def deltaStarsRaDecDeg2(ra1_deg, dec1_deg, ra2_deg, dec2_deg):
    ra1_rad = math.radians(ra1_deg)
    dec1_rad = math.radians(dec1_deg)
    ra2_rad = math.radians(ra2_deg)
    dec2_rad = math.radians(dec2_deg)

    sep_rad = math.acos(math.cos(90.0 - dec1_rad) * math.cos(90.0 - dec2_rad) +
                        math.sin(90.0 - dec1_rad) * math.sin(90.0 - dec2_rad) *
                        math.cos(ra1_rad - ra2_rad))
    res = math.degrees(sep_rad)
    return res


# Use spherical triangulation
deltaStarsRaDecDeg = deltaStarsRaDecDeg1
"""Use spherical triangulation."""


def get_starsep_RaDecDeg(ra1_deg, dec1_deg, ra2_deg, dec2_deg):
    """Calculate separation."""
    sep = deltaStarsRaDecDeg(ra1_deg, dec1_deg, ra2_deg, dec2_deg)
    sgn, deg, mn, sec = degToDms(sep)
    if deg != 0:
        txt = '%02d:%02d:%06.3f' % (deg, mn, sec)
    else:
        txt = '%02d:%06.3f' % (mn, sec)
    return txt


def add_offset_radec(ra_deg, dec_deg, delta_deg_ra, delta_deg_dec):
    """
    Algorithm to compute RA/Dec from RA/Dec base position plus tangent
    plane offsets.
    """
    # To radians
    x = math.radians(delta_deg_ra)
    y = math.radians(delta_deg_dec)
    raz = math.radians(ra_deg)
    decz = math.radians(dec_deg)

    sdecz = math.sin(decz)
    cdecz = math.cos(decz)

    d = cdecz - y * sdecz

    ra2 = math.atan2(x, d) + raz
    # Normalize ra into the range 0 to 2*pi
    twopi = math.pi * 2
    ra2 = math.fmod(ra2, twopi)
    if ra2 < 0.0:
        ra2 += twopi
    dec2 = math.atan2(sdecz + y * cdecz, math.sqrt(x * x + d * d))

    # back to degrees
    ra2_deg = math.degrees(ra2)
    dec2_deg = math.degrees(dec2)

    return (ra2_deg, dec2_deg)


def get_RaDecOffsets(ra1_deg, dec1_deg, ra2_deg, dec2_deg):
    """Calculate offset."""
    delta_ra_deg = ra1_deg - ra2_deg
    adj = math.cos(math.radians(dec2_deg))
    if delta_ra_deg > 180.0:
        delta_ra_deg = (delta_ra_deg - 360.0) * adj
    elif delta_ra_deg < -180.0:
        delta_ra_deg = (delta_ra_deg + 360.0) * adj
    else:
        delta_ra_deg *= adj

    delta_dec_deg = dec1_deg - dec2_deg
    return (delta_ra_deg, delta_dec_deg)


def calc_midpoint_radec(ra1, dec1, ra2, dec2):
    delta_ra = ra2 - ra1
    bx = np.cos(dec2) * np.cos(delta_ra)
    by = np.cos(dec2) * np.sin(delta_ra)
    dec_term1 = np.sin(dec1) + np.sin(dec2)
    dec_term2 = np.sqrt(np.power((np.cos(dec1) + bx), 2) +
                        np.power(by, 2))
    dec_mid = np.arctan2(dec_term1, dec_term2)
    ra_mid = ra1 + np.arctan2(by, np.cos(dec1) + bx)
    return (ra_mid, dec_mid)


def lon_to_deg(lon):
    """Convert longitude to degrees."""
    if isinstance(lon, str) and (':' in lon):
        # TODO: handle other coordinate systems
        lon_deg = hmsStrToDeg(lon)
    else:
        lon_deg = float(lon)
    return lon_deg


def lat_to_deg(lat):
    """Convert latitude to degrees."""
    if isinstance(lat, str) and (':' in lat):
        # TODO: handle other coordinate systems
        lat_deg = dmsStrToDeg(lat)
    else:
        lat_deg = float(lat)
    return lat_deg


def get_ruler_distances(image, p1, p2):
    """Get the distance calculated between two points.  A Bunch of
    results is returned, containing pixel values and distance values
    if the image contains a valid WCS.
    """
    x1, y1 = p1[:2]
    x2, y2 = p2[:2]

    dx, dy = x2 - x1, y2 - y1
    res = Bunch.Bunch(x1=x1, y1=y1, x2=x2, y2=y2,
                      theta=np.arctan2(y2 - y1, x2 - x1),
                      dx_pix=dx, dy_pix=dy,
                      dh_pix=np.sqrt(dx**2 + dy**2),
                      ra_org=None, dec_org=None,
                      ra_dst=None, dec_dst=None,
                      ra_heel=None, dec_heel=None,
                      dx_deg=None, dy_deg=None, dh_deg=None)

    if image is not None and hasattr(image, 'wcs') and image.wcs is not None:
        # Calculate RA and DEC for the three points
        try:
            # origination point
            ra_org, dec_org = image.pixtoradec(x1, y1)
            res.ra_org, res.dec_org = ra_org, dec_org

            # destination point
            ra_dst, dec_dst = image.pixtoradec(x2, y2)
            res.ra_dst, res.dec_dst = ra_dst, dec_dst

            # "heel" point making a right triangle
            ra_heel, dec_heel = image.pixtoradec(x2, y1)
            res.ra_heel, res.dec_heel = ra_heel, dec_heel

            res.dh_deg = deltaStarsRaDecDeg(ra_org, dec_org,
                                            ra_dst, dec_dst)
            res.dx_deg = deltaStarsRaDecDeg(ra_org, dec_org,
                                            ra_heel, dec_heel)
            res.dy_deg = deltaStarsRaDecDeg(ra_heel, dec_heel,
                                            ra_dst, dec_dst)
        except Exception as e:
            pass

    return res


def get_starsep_XY(image, x1, y1, x2, y2):
    # source point
    ra_org, dec_org = image.pixtoradec(x1, y1)

    # destination point
    ra_dst, dec_dst = image.pixtoradec(x2, y2)

    return get_starsep_RaDecDeg(ra_org, dec_org, ra_dst, dec_dst)


def calc_radius_xy(image, x, y, radius_deg):
    """Calculate a radius (in pixels) from the point (x, y) to a circle
    defined by radius in degrees.
    """
    # calculate ra/dec of x,y pixel
    ra_deg, dec_deg = image.pixtoradec(x, y)

    # Calculate position 1 degree from the given one
    # NOTE: this needs to add in DEC, not RA
    ra2_deg, dec2_deg = add_offset_radec(ra_deg, dec_deg,
                                         0.0, 1.0)

    # Calculate the length of this segment--it is pixels/deg
    x2, y2 = image.radectopix(ra2_deg, dec2_deg)
    px_per_deg_e = math.sqrt(math.fabs(x2 - x)**2 + math.fabs(y2 - y)**2)

    # calculate radius based on desired radius_deg
    radius_px = px_per_deg_e * radius_deg
    return radius_px


def calc_radius_deg2pix(image, ra_deg, dec_deg, delta_deg,
                        equinox=None):
    x, y = image.radectopix(ra_deg, dec_deg, equinox=equinox)
    return calc_radius_xy(image, x, y, delta_deg)


def add_offset_xy(image, x, y, delta_deg_x, delta_deg_y):
    # calculate ra/dec of x,y pixel
    ra_deg, dec_deg = image.pixtoradec(x, y)

    # add offsets
    ra2_deg, dec2_deg = add_offset_radec(ra_deg, dec_deg,
                                         delta_deg_x, delta_deg_y)

    # then back to new pixel coords
    x2, y2 = image.radectopix(ra2_deg, dec2_deg)

    return (x2, y2)


def calc_radius_center(image, delta_deg):
    return calc_radius_xy(image,
                          float(image.width / 2.0),
                          float(image.height / 2.0),
                          delta_deg)


def calc_compass(image, x, y, len_deg_e, len_deg_n):

    # Get east and north coordinates
    xe, ye = add_offset_xy(image, x, y, len_deg_e, 0.0)
    xn, yn = add_offset_xy(image, x, y, 0.0, len_deg_n)

    return (x, y, xn, yn, xe, ye)


def calc_compass_radius(image, x, y, radius_px):
    xe, ye = add_offset_xy(image, x, y, 1.0, 0.0)
    xn, yn = add_offset_xy(image, x, y, 0.0, 1.0)

    # now calculate the length in pixels of those arcs
    # (planar geometry is good enough here)
    px_per_deg_e = math.sqrt(math.fabs(ye - y)**2 + math.fabs(xe - x)**2)
    px_per_deg_n = math.sqrt(math.fabs(yn - y)**2 + math.fabs(xn - x)**2)

    # now calculate the arm length in degrees for each arm
    # (this produces same-length arms)
    len_deg_e = radius_px / px_per_deg_e
    len_deg_n = radius_px / px_per_deg_n

    return calc_compass(image, x, y, len_deg_e, len_deg_n)


def calc_compass_center(image):
    # calculate center of data
    x = float(image.width) / 2.0
    y = float(image.height) / 2.0

    # radius we want the arms to be (approx 1/4 the smallest dimension)
    radius_px = float(min(image.width, image.height)) / 4.0

    return calc_compass_radius(image, x, y, radius_px)


# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``Errors`` plugin reports error messages on the viewer.

**Plugin Type: Global**

``Errors`` is a global plugin.  Only one instance can be opened.

**Usage**

When an error occurs in Ginga, its message may be reported here.

This plugin is not usually configured to be closeable, but the user can
make it so by setting the "closeable" setting to True in the configuration
file--then Close and Help buttons will be added to the bottom of the UI.

"""
import time
from collections import deque

from ginga import GingaPlugin
from ginga.gw import Widgets

__all__ = ['Errors']


class Errors(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(Errors, self).__init__(fv)

        spec = self.fv.get_plugin_spec(str(self))

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Errors')
        self.settings.add_defaults(closeable=not spec.get('hidden', False),
                                   max_errors=100)
        self.settings.load(onError='silent')

        max_errors = self.settings.get('max_errors', 100)
        self.pending_errors = deque([], max_errors)
        self.gui_up = False

    def build_gui(self, container):
        self.msg_font = self.fv.get_font('fixed', 10)

        vbox = Widgets.VBox()

        mlst = Widgets.VBox()
        mlst.set_spacing(2)
        self.msg_list = mlst

        sw = Widgets.ScrollArea()
        sw.set_widget(self.msg_list)

        vbox.add_widget(sw, stretch=1)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(4)

        if self.settings.get('closeable', False):
            btn = Widgets.Button("Close")
            btn.add_callback('activated', lambda w: self.close())
            btns.add_widget(btn)
            btn = Widgets.Button("Help")
            btn.add_callback('activated', lambda w: self.help())
            btns.add_widget(btn, stretch=0)

        btn = Widgets.Button("Remove All")
        btn.add_callback('activated', lambda w: self.remove_all())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)

        vbox.add_widget(btns, stretch=0)
        container.add_widget(vbox, stretch=1)

        self.gui_up = True

    def add_error(self, errmsg, ts=None):
        if ts is None:
            # Add the time the error occurred
            ts = time.strftime("%m/%d %H:%M:%S", time.localtime())

        if not self.gui_up:
            self.pending_errors.append((errmsg, ts))
            return

        vbox = Widgets.VBox()

        hbox = Widgets.HBox()
        # Add the time the error occurred
        ts = time.strftime("%m/%d %H:%M:%S", time.localtime())
        lbl = Widgets.Label(ts, halign='left')
        hbox.add_widget(lbl, stretch=0)
        hbox.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(hbox, stretch=0)

        tw = Widgets.TextArea(editable=False, wrap=False)
        tw.set_font(self.msg_font)

        tw.set_text(errmsg)
        vbox.add_widget(tw, stretch=1)

        hbox = Widgets.HBox()
        btn = Widgets.Button("Remove")
        btn.add_callback('activated', lambda w: self.remove_error(vbox))
        hbox.add_widget(btn)
        hbox.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(hbox, stretch=0)
        # special hack for Qt
        vbox.cfg_expand(horizontal=1)

        self.msg_list.add_widget(vbox, stretch=0)
        # TODO: force scroll to bottom

    def remove_error(self, child):
        self.msg_list.remove(child)

    def remove_all(self):
        self.pending_errors.clear()
        for child in list(self.msg_list.get_children()):
            self.remove_error(child)

    def start(self):
        pending = self.pending_errors
        self.pending_errors = []

        for errmsg, ts in pending:
            self.add_error(errmsg, ts=ts)

    def stop(self):
        self.pending_errors = []
        self.gui_up = False

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'errors'

# END



#
# ImageViewTk.py -- a backend for Ginga using a Tk canvas widget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import PIL.Image as PILimage

have_pil_imagetk = False
try:
    # sometimes this is not installed even though PIL is
    from PIL.ImageTk import PhotoImage
    have_pil_imagetk = True
except ImportError:
    from tkinter import PhotoImage

from ginga import ImageView, Mixins, Bindings  # noqa
from ginga.canvas.mixins import DrawingMixin, CanvasMixin, CompoundMixin  # noqa
from ginga.canvas import render
from ginga.util.toolbox import ModeIndicator  # noqa

from . import TkHelp  # noqa


class ImageViewTkError(ImageView.ImageViewError):
    pass


class ImageViewTk(ImageView.ImageViewBase):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageView.ImageViewBase.__init__(self, logger=logger,
                                         rgbmap=rgbmap,
                                         settings=settings)

        self.tkcanvas = None
        self.tkphoto = None

        self.t_.set_defaults(renderer='cairo')

        self._defer_task = None
        self.msgtask = None

        self.rgb_order = 'RGBA'

        self.renderer = None
        # Pick a renderer that can work with us
        renderers = ['cairo', 'agg', 'pil', 'opencv']
        preferred = self.t_['renderer']
        if preferred in renderers:
            renderers.remove(preferred)
        self.possible_renderers = [preferred] + renderers
        self.choose_best_renderer()

    def set_widget(self, canvas):
        """Call this method with the Tkinter canvas that will be used
        for the display.
        """
        self.tkcanvas = canvas

        canvas.bind("<Configure>", self._resize_cb)
        width = canvas.winfo_width()
        height = canvas.winfo_height()

        # see reschedule_redraw() method
        self._defer_task = TkHelp.Timer(tkcanvas=canvas)
        self._defer_task.add_callback('expired',
                                      lambda timer: self.delayed_redraw())
        self.msgtask = TkHelp.Timer(tkcanvas=canvas)
        self.msgtask.add_callback('expired',
                                  lambda timer: self.onscreen_message(None))

        self.configure_window(width, height)

    def get_widget(self):
        return self.tkcanvas

    def choose_renderer(self, name):
        klass = render.get_render_class(name)
        self.renderer = klass(self)

        if self.tkcanvas is not None:
            wd = self.tkcanvas.winfo_width()
            ht = self.tkcanvas.winfo_height()
            self.configure_window(wd, ht)

    def choose_best_renderer(self):
        for name in self.possible_renderers:
            try:
                self.choose_renderer(name)
                self.logger.info("best renderer available is '{}'".format(name))
                return
            except Exception as e:
                continue

        raise ImageViewTkError("No valid renderers available: {}".format(str(self.possible_renderers)))

    def update_image(self):
        if self.tkcanvas is None:
            return

        cr = self.tkcanvas

        # remove all old items from the canvas
        items = cr.find_all()
        for item in items:
            cr.delete(item)

        wd, ht = self.get_window_size()

        # make a Tk photo image and stick it to the canvas

        if have_pil_imagetk:
            # Get surface as a numpy array
            arr8 = self.renderer.get_surface_as_array(order='RGB')
            image = PILimage.fromarray(arr8)
            photo = PhotoImage(image)

        else:
            # fallback to a little slower method--make a PNG image
            image = self.get_surface_as_rgb_format_bytes(format='png')
            photo = PhotoImage(data=image)

        # hang on to a reference otherwise it gets gc'd
        self.tkphoto = photo

        cr.create_image(0, 0, anchor='nw', image=photo)

        # is this necessary?
        cr.config(scrollregion=cr.bbox('all'))

    def reschedule_redraw(self, time_sec):
        self._defer_task.stop()
        self._defer_task.start(time_sec)

    def configure_window(self, width, height):
        self.configure_surface(width, height)

    def _resize_cb(self, event):
        self.configure_window(event.width, event.height)

    def set_cursor(self, cursor):
        if self.tkcanvas is None:
            return
        self.tkcanvas.config(cursor=cursor)

    def onscreen_message(self, text, delay=None, redraw=True):
        if self.tkcanvas is None:
            return
        self.msgtask.stop()
        self.set_onscreen_message(text, redraw=redraw)
        if delay is not None:
            self.msgtask.start(delay)

    def take_focus(self):
        self.tkcanvas.focus_set()


class ImageViewEvent(ImageViewTk):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageViewTk.__init__(self, logger=logger, rgbmap=rgbmap,
                             settings=settings)

        self._button = 0

        # @$%&^(_)*&^ tk!!
        self._keytbl = {
            'shift_l': 'shift_l',
            'shift_r': 'shift_r',
            'control_l': 'control_l',
            'control_r': 'control_r',
            'alt_l': 'alt_l',
            'alt_r': 'alt_r',
            'super_l': 'super_l',
            'super_r': 'super_r',
            'meta_right': 'meta_right',
            'asciitilde': '~',
            'grave': 'backquote',
            'exclam': '!',
            'at': '@',
            'numbersign': '#',
            'percent': '%',
            'asciicircum': '^',
            'ampersand': '&',
            'asterisk': '*',
            'dollar': '$',
            'parenleft': '(',
            'parenright': ')',
            'underscore': '_',
            'minus': '-',
            'plus': '+',
            'equal': '=',
            'braceleft': '{',
            'braceright': '}',
            'bracketleft': '[',
            'bracketright': ']',
            'bar': '|',
            'colon': ':',
            'semicolon': ';',
            'quotedbl': 'doublequote',
            'apostrophe': 'singlequote',
            'backslash': 'backslash',
            'less': '<',
            'greater': '>',
            'comma': ',',
            'period': '.',
            'question': '?',
            'slash': '/',
            'space': 'space',
            'escape': 'escape',
            'return': 'return',
            'tab': 'tab',
            'f1': 'f1',
            'f2': 'f2',
            'f3': 'f3',
            'f4': 'f4',
            'f5': 'f5',
            'f6': 'f6',
            'f7': 'f7',
            'f8': 'f8',
            'f9': 'f9',
            'f10': 'f10',
            'f11': 'f11',
            'f12': 'f12',
            'right': 'right',
            'left': 'left',
            'up': 'up',
            'down': 'down',
            'insert': 'insert',
            'delete': 'delete',
            'home': 'home',
            'end': 'end',
            'prior': 'page_up',
            'next': 'page_down',
        }

        # Define cursors for pick and pan
        #hand = openHandCursor()
        hand = 'fleur'
        self.define_cursor('pan', hand)
        cross = 'cross'
        self.define_cursor('pick', cross)

        for name in ('motion', 'button-press', 'button-release',
                     'key-press', 'key-release', 'drag-drop',
                     'scroll', 'map', 'focus', 'enter', 'leave',
                     ):
            self.enable_callback(name)

    def set_widget(self, canvas):
        super(ImageViewEvent, self).set_widget(canvas)

        canvas.bind("<Enter>", self.enter_notify_event)
        canvas.bind("<Leave>", self.leave_notify_event)
        canvas.bind("<FocusIn>", lambda evt: self.focus_event(evt, True))
        canvas.bind("<FocusOut>", lambda evt: self.focus_event(evt, False))
        canvas.bind("<KeyPress>", self.key_press_event)
        canvas.bind("<KeyRelease>", self.key_release_event)
        #canvas.bind("<Map>", self.map_event)
        # scroll events in tk are overloaded into the button press events
        canvas.bind("<ButtonPress>", self.button_press_event)
        canvas.bind("<ButtonRelease>", self.button_release_event)
        canvas.bind("<Motion>", self.motion_notify_event)

        # TODO: Set up widget as a drag and drop destination

        return self.make_callback('map')

    def transkey(self, keyname):
        self.logger.debug("key name in tk '%s'" % (keyname))
        try:
            return self._keytbl[keyname.lower()]

        except KeyError:
            return keyname

    def get_key_table(self):
        return self._keytbl

    def focus_event(self, event, hasFocus):
        return self.make_callback('focus', hasFocus)

    def enter_notify_event(self, event):
        # Does widget accept focus when mouse enters window
        enter_focus = self.t_.get('enter_focus', False)
        if enter_focus:
            self.tkcanvas.focus_set()
        return self.make_callback('enter')

    def leave_notify_event(self, event):
        self.logger.debug("leaving widget...")
        return self.make_callback('leave')

    def key_press_event(self, event):
        # without this we do not get key release events if the focus
        # changes to another window
        self.tkcanvas.grab_set_global()

        keyname = event.keysym
        keyname = self.transkey(keyname)
        self.logger.debug("key press event, key=%s" % (keyname))
        return self.make_ui_callback('key-press', keyname)

    def key_release_event(self, event):
        self.tkcanvas.grab_release()

        keyname = event.keysym
        keyname = self.transkey(keyname)
        self.logger.debug("key release event, key=%s" % (keyname))
        return self.make_ui_callback('key-release', keyname)

    def button_press_event(self, event):
        x = event.x
        y = event.y
        self.last_win_x, self.last_win_y = x, y

        button = 0
        if event.num != 0:
            # some kind of wierd convention for scrolling, shoehorned into
            # Tk, I guess
            if event.num in (4, 5):
                direction = 0.0   # up
                if event.num == 5:
                    # down
                    direction = 180.0
                # 15 deg is standard 1-click turn for a wheel mouse
                num_degrees = 15.0
                self.logger.debug("scroll deg=%f direction=%f" % (
                    num_degrees, direction))

                data_x, data_y = self.check_cursor_location()

                return self.make_ui_callback('scroll', direction, num_degrees,
                                             data_x, data_y)

            button |= 0x1 << (event.num - 1)
        self._button = button
        self.logger.debug("button event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('button-press', button, data_x, data_y)

    def button_release_event(self, event):
        # event.button, event.x, event.y
        x = event.x
        y = event.y
        self.last_win_x, self.last_win_y = x, y

        button = 0
        if event.num != 0:
            if event.num in (4, 5):
                return False

            button |= 0x1 << (event.num - 1)
        self._button = 0
        self.logger.debug("button release at %dx%d button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('button-release', button, data_x, data_y)

    def motion_notify_event(self, event):
        #button = 0
        button = self._button
        x, y = event.x, event.y
        self.last_win_x, self.last_win_y = x, y

        # num = event.num
        # if num == 1:
        #     button |= 0x1
        # elif num == 2:
        #     button |= 0x2
        # elif num == 3:
        #     button |= 0x4
        self.logger.debug("motion event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('motion', button, data_x, data_y)

    ## def drop_event(self, widget, context, x, y, selection, targetType,
    ##                time):
    ##     if targetType != self.TARGET_TYPE_TEXT:
    ##         return False
    ##     paths = selection.data.split('\n')
    ##     self.logger.debug("dropped filename(s): %s" % (str(paths)))
    ##     return self.make_ui_callback('drag-drop', paths)


class ImageViewZoom(Mixins.UIMixin, ImageViewEvent):

    # class variables for binding map and bindings can be set
    bindmapClass = Bindings.BindingMapper
    bindingsClass = Bindings.ImageViewBindings

    @classmethod
    def set_bindingsClass(cls, klass):
        cls.bindingsClass = klass

    @classmethod
    def set_bindmapClass(cls, klass):
        cls.bindmapClass = klass

    def __init__(self, logger=None, rgbmap=None, settings=None,
                 bindmap=None, bindings=None):
        ImageViewEvent.__init__(self, logger=logger, rgbmap=rgbmap,
                                settings=settings)
        Mixins.UIMixin.__init__(self)

        self.ui_set_active(True)

        if bindmap is None:
            bindmap = ImageViewZoom.bindmapClass(self.logger)
        self.bindmap = bindmap
        bindmap.register_for_events(self)

        if bindings is None:
            bindings = ImageViewZoom.bindingsClass(self.logger)
        self.set_bindings(bindings)

    def get_bindmap(self):
        return self.bindmap

    def get_bindings(self):
        return self.bindings

    def set_bindings(self, bindings):
        self.bindings = bindings
        bindings.set_bindings(self)


class CanvasView(ImageViewZoom):

    def __init__(self, logger=None, settings=None, rgbmap=None,
                 bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger, settings=settings,
                               rgbmap=rgbmap,
                               bindmap=bindmap, bindings=bindings)

        # Needed for UIMixin to propagate events correctly
        self.objects = [self.private_canvas]

    def set_canvas(self, canvas, private_canvas=None):
        super(CanvasView, self).set_canvas(canvas,
                                           private_canvas=private_canvas)

        self.objects[0] = self.private_canvas


class ImageViewCanvasError(ImageViewTkError):
    pass


class ImageViewCanvas(ImageViewZoom,
                      DrawingMixin, CanvasMixin, CompoundMixin):

    def __init__(self, logger=None, rgbmap=None, settings=None,
                 bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger,
                               rgbmap=rgbmap,
                               settings=settings,
                               bindmap=bindmap,
                               bindings=bindings)
        CompoundMixin.__init__(self)
        CanvasMixin.__init__(self)
        DrawingMixin.__init__(self)

        # we are both a viewer and a canvas
        self.set_canvas(self, private_canvas=self)

        self._mi = ModeIndicator(self)

#END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``RC`` plugin implements a remote control interface for the Ginga
viewer.

**Plugin Type: Global**

``RC`` is a global plugin.  Only one instance can be opened.

**Usage**

The ``RC`` (Remote Control) plugin provides a way to control Ginga remotely
through the use of an XML-RPC interface.  Start the plugin from the
"Plugins" menu (invoke "Start RC") or launch ginga with the ``--modules=RC``
command line option to start it automatically.

By default, the plugin starts up with server running on port 9000 bound
to the localhost interface -- this allows connections only from the local
host.  If you want to change this, set the host and port in the "Set
Addr" control and press ``Enter`` -- you should see the address update in the
"Addr:" display field.

Please note that the host part (before the colon) does not indicate
*which* host you want to allow access from, but to which interface to
bind.  If you want to allow any host to connect, leave it blank (but
include the colon and port number) to allow the server to bind on all
interfaces. Press "Restart" to then restart the server at the new
address.

Once the plugin is started, you can use the ``ggrc`` script (included when
``ginga`` is installed) to control Ginga.  Take a look at the script if you
want to see how to write your own programmatic interface.

Show example usage::

        $ ggrc help

Show help for a specific Ginga method::

        $ ggrc help ginga <method>

Show help for a specific channel method::

        $ ggrc help channel <chname> <method>

Ginga (viewer shell) methods can be called like this::

        $ ggrc ginga <method> <arg1> <arg2> ...

Per-channel methods can be called like this::

        $ ggrc channel <chname> <method> <arg1> <arg2> ...

Calls can be made from a remote host by adding the options::

        --host=<hostname> --port=9000

(In the plugin GUI, be sure to remove the "localhost" prefix
from the "addr", but leave the colon and port.)

**Examples**

Create a new channel::

        $ ggrc ginga add_channel FOO

Load a file::

        $ ggrc ginga load_file /home/eric/testdata/SPCAM/SUPA01118797.fits

Load a file into a specific channel::

        $ ggrc ginga load_file /home/eric/testdata/SPCAM/SUPA01118797.fits FOO

Cut levels::

        $ ggrc channel FOO cut_levels 163 1300

Auto cut levels::

        $ ggrc channel FOO auto_levels

Zoom to a specific level::

        $ ggrc -- channel FOO zoom_to -7

(Note the use of ``--`` to allow us to pass a parameter beginning with ``-``.)

Zoom to fit::

        $ ggrc channel FOO zoom_fit

Transform (arguments are a boolean triplet: ``flipx`` ``flipy`` ``swapxy``)::

        $ ggrc channel FOO transform 1 0 1

Rotate::

        $ ggrc channel FOO rotate 37.5

Change colormap::

        $ ggrc channel FOO set_color_map rainbow3

Change color distribution algorithm::

        $ ggrc channel FOO set_color_algorithm log

Change intensity map::

        $ ggrc channel FOO set_intensity_map neg

In some cases, you may need to resort to shell escapes to be able to
pass certain characters to Ginga.  For example, a leading dash character is
usually interpreted as a program option.  In order to pass a signed
integer, you may need to do something like::

        $ ggrc -- channel FOO zoom -7

**Interfacing from within Python**

It is also possible to control Ginga in RC mode from within Python.
The following describes some of the functionality.

*Connecting*

First, launch Ginga and start the ``RC`` plugin.
This can be done from the command line::

        ginga --modules=RC

From within Python, connect with a ``RemoteClient`` object as
follows::

        from ginga.util import grc
        host='localhost'
        port=9000
        viewer = grc.RemoteClient(host, port)

This viewer object is now linked to the Ginga using ``RC``.

*Load an Image*

You can load an image from memory in a channel of
your choosing.  First, connect to a channel::

        ch = viewer.channel('Image')

Then, load a Numpy image (i.e., any 2D ``ndarray``)::

        import numpy as np
        img = np.random.rand(500, 500) * 10000.0
        ch.load_np('Image_Name', img, 'fits', {})

The image will display in Ginga and can be manipulated
as usual.

*Overlay a Canvas Object*

It is possible to add objects to the canvas in a given
channel.  First, connect::

        canvas = viewer.canvas('Image')

This connects to the channel named "Image".  You can
clear the objects drawn in the canvas::

        canvas.clear()

You can also add any basic canvas object.  The key issue to keep in
mind is that the objects input must pass through the XMLRC
protocol.  This means simple data types (``float``, ``int``, ``list``,
or ``str``); No arrays.  Here is an example to plot a line through a series
of points defined by two Numpy arrays::

        x = np.arange(100)
        y = np.sqrt(x)
        points = list(zip(x.tolist(), y.tolist()))
        canvas.add('path', points, color='red')

This will draw a red line on the image.

"""
import sys
import bz2
from io import BytesIO

import numpy as np

from ginga import GingaPlugin
from ginga import AstroImage
from ginga.gw import Widgets
from ginga.util import grc

__all__ = ['RC']

help_msg = sys.modules[__name__].__doc__


class RC(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(RC, self).__init__(fv)

        # What port to listen for requests
        self.port = 9000
        # If blank, listens on all interfaces
        self.host = 'localhost'

        # this will hold the remote object
        self.robj = None
        # this will hold the remote object server
        self.server = None

        self.ev_quit = fv.ev_quit

    def build_gui(self, container):
        vbox = Widgets.VBox()

        fr = Widgets.Frame("Remote Control")

        captions = [
            ("Addr:", 'label', "Addr", 'llabel', 'Restart', 'button'),
            ("Set Addr:", 'label', "Set Addr", 'entry')]
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        addr = self.host + ':' + str(self.port)
        b.addr.set_text(addr)
        b.restart.set_tooltip("Restart the server")
        b.restart.add_callback('activated', self.restart_cb)

        b.set_addr.set_length(100)
        b.set_addr.set_text(addr)
        b.set_addr.set_tooltip("Set address to run remote control server")
        b.set_addr.add_callback('activated', self.set_addr_cb)

        fr.set_widget(w)
        vbox.add_widget(fr, stretch=0)

        # stretch
        vbox.add_widget(Widgets.Label(''), stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(4)
        btns.set_border_width(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(btns)

        container.add_widget(vbox, stretch=1)

    def start(self):
        self.robj = GingaWrapper(self.fv, self.logger)

        self.server = grc.RemoteServer(self.robj,
                                       host=self.host, port=self.port,
                                       ev_quit=self.fv.ev_quit,
                                       logger=self.logger)
        self.server.start(thread_pool=self.fv.get_threadPool())

    def stop(self):
        self.server.stop()

    def restart_cb(self, w):
        # restart server
        self.server.stop()
        self.start()

    def set_addr_cb(self, w):
        # get and parse address
        addr = w.get_text()
        host, port = addr.split(':')
        self.host = host
        self.port = int(port)
        self.w.addr.set_text(addr)

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def __str__(self):
        return 'rc'


class GingaWrapper(object):

    def __init__(self, fv, logger):
        self.fv = fv
        self.logger = logger

    def help(self, *args):
        """Get help for a remote interface method.

        Examples
        --------
        help('ginga', `method`)
            name of the method for which you want help

        help('channel', `chname`, `method`)
            name of the method in the channel for which you want help

        Returns
        -------
        help : string
            a help message
        """
        if len(args) == 0:
            return help_msg

        which = args[0].lower()

        if which == 'ginga':
            method = args[1]
            _method = getattr(self.fv, method)
            return _method.__doc__

        elif which == 'channel':
            chname = args[1]
            method = args[2]
            chinfo = self.fv.get_channel(chname)
            _method = getattr(chinfo.viewer, method)
            return _method.__doc__

        else:
            return ("Please use 'help ginga <method>' or "
                    "'help channel <chname> <method>'")

    def load_buffer(self, imname, chname, img_buf, dims, dtype,
                    header, metadata, compressed):
        """Display a FITS image buffer.

        Parameters
        ----------
        imname : string
            a name to use for the image in Ginga
        chname : string
            channel in which to load the image
        img_buf : string
            the image data, as a bytes object
        dims : tuple
            image dimensions in pixels (usually (height, width))
        dtype : string
            numpy data type of encoding (e.g. 'float32')
        header : dict
            fits file header as a dictionary
        metadata : dict
            other metadata about image to attach to image
        compressed : bool
            decompress buffer using "bz2"

        Returns
        -------
        0

        Notes
        -----

        * Get array dims: data.shape
        * Get array dtype: str(data.dtype)
        * Make a string from a numpy array: buf = data.tostring()
        * Compress the buffer: buf = bz2.compress(buf)

        """
        self.logger.info("received image data len=%d" % (len(img_buf)))

        # Unpack the data
        try:
            # Uncompress data if necessary
            decompress = metadata.get('decompress', None)
            if compressed or (decompress == 'bz2'):
                img_buf = bz2.decompress(img_buf)

            # dtype string works for most instances
            if dtype == '':
                dtype = np.float

            byteswap = metadata.get('byteswap', False)

            # Create image container
            image = AstroImage.AstroImage(logger=self.logger)
            image.load_buffer(img_buf, dims, dtype, byteswap=byteswap,
                              metadata=metadata)
            image.update_keywords(header)
            image.set(name=imname, path=None)

        except Exception as e:
            # Some kind of error unpacking the data
            errmsg = "Error creating image data for '%s': %s" % (
                imname, str(e))
            self.logger.error(errmsg)
            raise GingaPlugin.PluginError(errmsg)

        # Display the image
        channel = self.fv.gui_call(self.fv.get_channel_on_demand, chname)

        # Note: this little hack needed to let window resize in time for
        # file to auto-size properly
        self.fv.gui_do(self.fv.change_channel, channel.name)

        self.fv.gui_do(self.fv.add_image, imname, image,
                       chname=channel.name)
        return 0

    def load_fits_buffer(self, imname, chname, file_buf, num_hdu,
                         metadata):
        from astropy.io import fits

        # Unpack the data
        try:
            # Uncompress data if necessary
            decompress = metadata.get('decompress', None)
            if decompress == 'bz2':
                file_buf = bz2.decompress(file_buf)

            self.logger.info("received data: len=%d num_hdu=%d" % (
                len(file_buf), num_hdu))

            in_f = BytesIO(file_buf)

            # Create image container
            with fits.open(in_f, 'readonly') as fits_f:
                image = AstroImage.AstroImage(metadata=metadata,
                                              logger=self.logger)
                image.load_hdu(fits_f[num_hdu], fobj=fits_f)
            image.set(name=imname, path=None, idx=num_hdu)

        except Exception as e:
            # Some kind of error unpacking the data
            errmsg = "Error creating image data for '%s': %s" % (
                imname, str(e))
            self.logger.error(errmsg)
            raise GingaPlugin.PluginError(errmsg)

        # Display the image
        channel = self.fv.gui_call(self.fv.get_channel_on_demand, chname)

        # Note: this little hack needed to let window resize in time for
        # file to auto-size properly
        self.fv.gui_do(self.fv.change_channel, channel.name)

        self.fv.gui_do(self.fv.add_image, imname, image,
                       chname=channel.name)
        return 0

    def channel(self, chname, method_name, *args, **kwdargs):
        chinfo = self.fv.get_channel(chname)
        _method = getattr(chinfo.viewer, method_name)
        return self.fv.gui_call(_method, *args, **kwdargs)

    def ginga(self, method_name, *args, **kwdargs):
        _method = getattr(self.fv, method_name)
        return self.fv.gui_call(_method, *args, **kwdargs)

    def canvas(self, chname, command, *args, **kwargs):
        chinfo = self.fv.get_channel(chname)
        canvas = chinfo.viewer.get_canvas()
        if command == 'add':
            klass = canvas.get_draw_class(args[0])
            newargs = args[1:]
            obj = klass(*newargs, **kwargs)
            return self.fv.gui_call(canvas.add, obj)

        elif command == 'clear':
            # Clear only drawn objects, not the image
            nobj = len(canvas.objects)
            if nobj == 0:
                return
            elif nobj == 1:  # Assume it is the image, don't remove
                return
            else:
                canvas.objects = canvas.objects[0:1]

        elif command == 'nobj':
            return len(canvas.objects)

        else:
            print("Canvas RC command not recognized")
            return

# END



#
# CanvasRenderPil.py -- for rendering into a Ginga widget with pillow
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import math
import numpy as np

from PIL import Image

from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc

from . import PilHelp


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)

        # TODO: encapsulate this drawable
        self.cr = PilHelp.PilContext(surface)

        self.pen = None
        self.brush = None
        self.font = None

    def set_line_from_shape(self, shape):
        # TODO: support line width and style
        alpha = getattr(shape, 'alpha', 1.0)
        self.pen = self.cr.get_pen(shape.color, alpha=alpha)

    def set_fill_from_shape(self, shape):
        fill = getattr(shape, 'fill', False)
        if fill:
            if hasattr(shape, 'fillcolor') and shape.fillcolor:
                color = shape.fillcolor
            else:
                color = shape.color
            alpha = getattr(shape, 'alpha', 1.0)
            alpha = getattr(shape, 'fillalpha', alpha)
            self.brush = self.cr.get_brush(color, alpha=alpha)
        else:
            self.brush = None

    def set_font_from_shape(self, shape):
        if hasattr(shape, 'font'):
            if (hasattr(shape, 'fontsize') and shape.fontsize is not None and
                not getattr(shape, 'fontscale', False)):
                fontsize = shape.fontsize
            else:
                fontsize = shape.scale_font(self.viewer)
            alpha = getattr(shape, 'alpha', 1.0)
            fontsize = self.scale_fontsize(fontsize)
            self.font = self.cr.get_font(shape.font, fontsize, shape.color,
                                         alpha=alpha)
        else:
            self.font = None

    def initialize_from_shape(self, shape, line=True, fill=True, font=True):
        if line:
            self.set_line_from_shape(shape)
        if fill:
            self.set_fill_from_shape(shape)
        if font:
            self.set_font_from_shape(shape)

    def set_line(self, color, alpha=1.0, linewidth=1, style='solid'):
        # TODO: support line width and style
        self.pen = self.cr.get_pen(color, alpha=alpha)

    def set_fill(self, color, alpha=1.0):
        if color is None:
            self.brush = None
        else:
            self.brush = self.cr.get_brush(color, alpha=alpha)

    def set_font(self, fontname, fontsize, color='black', alpha=1.0):
        fontsize = self.scale_fontsize(fontsize)
        self.font = self.cr.get_font(fontname, fontsize, color,
                                     alpha=alpha)

    def text_extents(self, text):
        return self.cr.text_extents(text, self.font)

    def get_affine_transform(self, cx, cy, rot_deg):
        x, y = 0, 0          # old center
        nx, ny = cx, cy      # new center
        sx = sy = 1.0        # new scale
        cosine = math.cos(math.radians(rot_deg))
        sine = math.sin(math.radians(rot_deg))
        a = cosine / sx
        b = sine / sx
        c = x - nx * a - ny * b
        d = -sine / sy
        e = cosine / sy
        f = y - nx * d - ny * e
        return (a, b, c, d, e, f)

    ##### DRAWING OPERATIONS #####

    def draw_text(self, cx, cy, text, rot_deg=0.0):
        wd, ht = self.cr.text_extents(text, self.font)

        # NOTE: rotation ignored in PIL, for now
        self.cr.text((cx, cy - ht), text, self.font, self.pen)

    def draw_polygon(self, cpoints):
        cpoints = trcalc.strip_z(cpoints)
        self.cr.polygon(cpoints, self.pen, self.brush)

    def draw_circle(self, cx, cy, cradius):
        self.cr.circle((cx, cy), cradius, self.pen, self.brush)

    def draw_line(self, cx1, cy1, cx2, cy2):
        self.cr.line((cx1, cy1), (cx2, cy2), self.pen)

    def draw_path(self, cpoints):
        cpoints = trcalc.strip_z(cpoints)
        self.cr.path(cpoints, self.pen)


class CanvasRenderer(render.RendererBase):

    def __init__(self, viewer):
        render.RendererBase.__init__(self, viewer)

        self.kind = 'pil'
        self.rgb_order = 'RGBA'
        self.surface = None
        self.dims = ()

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        width, height = dims[:2]
        self.dims = (width, height)
        self.logger.debug("renderer reconfigured to %dx%d" % (
            width, height))

        # create PIL surface the size of the window
        # NOTE: pillow needs an RGB surface in order to draw with alpha
        # blending, not RGBA
        self.surface = Image.new('RGB', (width, height), color=0)

    def render_image(self, rgbobj, dst_x, dst_y):
        """Render the image represented by (rgbobj) at dst_x, dst_y
        in the pixel space.
        *** internal method-- do not use ***
        """
        if self.surface is None:
            return
        self.logger.debug("redraw surface")

        # get window contents as a buffer and paste it into the PIL surface
        # TODO: allow greater bit depths when support is better in PIL
        rgb_arr = self.viewer.getwin_array(order=self.rgb_order, dtype=np.uint8)
        p_image = Image.fromarray(rgb_arr)

        if self.surface is None or p_image.size != self.surface.size:
            # window size must have changed out from underneath us!
            width, height = self.viewer.get_window_size()
            self.resize((width, height))
            if p_image.size != self.surface.size:
                raise render.RenderError("Rendered image does not match window size")

        self.surface.paste(p_image)

    def get_surface_as_array(self, order=None):
        if self.surface is None:
            raise render.RenderError("No PIL surface defined")

        # TODO: could these have changed between the time that self.surface
        # was last updated?
        wd, ht = self.dims[:2]

        # Get agg surface as a numpy array
        arr8 = np.fromstring(self.surface.tobytes(), dtype=np.uint8)
        arr8 = arr8.reshape((ht, wd, 3))

        # adjust according to viewer's needed order
        arr8 = self.reorder(order, arr8, 'RGB')
        return arr8

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.surface)
        cr.initialize_from_shape(shape, font=False)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        cr.set_font_from_shape(shape)
        return cr.text_extents(shape.text)

#END



"""Unit Tests for the Future class"""

import sys
import threading
import time

import pytest

from ginga.misc.Future import Future, TimeoutError


class _TestError(Exception):
    pass


class TestFuture(object):

    def setup_class(self):
        self.future_thread = None

    def test_init(self):
        test_future = Future()

        assert hasattr(test_future, 'cb')

        if sys.version_info.major == 2:
            assert isinstance(test_future.evt, threading._Event)
        elif sys.version_info.major == 3:
            assert isinstance(test_future.evt, threading.Event)

        assert test_future.evt.is_set() is False
        assert test_future.res is None
        assert test_future.data is None
        assert 'resolved' in test_future.cb

        expected = []
        actual = test_future.cb['resolved']
        assert expected == actual

    def test_init_with_data(self):
        test_future = Future("TestData")

        assert hasattr(test_future, 'cb')

        if sys.version_info.major == 2:
            assert isinstance(test_future.evt, threading._Event)
        elif sys.version_info.major == 3:
            assert isinstance(test_future.evt, threading.Event)

        assert test_future.evt.is_set() is False
        assert test_future.res is None
        assert test_future.data == "TestData"
        assert 'resolved' in test_future.cb

        expected = []
        actual = test_future.cb['resolved']
        assert expected == actual

    def test_get_data_no_data(self):
        test_future = Future()

        expected = None
        actual = test_future.get_data()
        assert expected == actual

    def test_get_data_some_data(self):
        test_future = Future("TestData")

        expected = "TestData"
        actual = test_future.get_data()
        assert expected == actual

    def test_freeze(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            pass

        test_future.freeze(
            test_method, "arg1", "arg2", kwarg1="test", kwarg2="test")

        assert test_future.method == test_method
        assert test_future.args == ("arg1", "arg2")
        assert test_future.kwdargs == {"kwarg1": "test", "kwarg2": "test"}

    def test_freeze_empty_args(self):
        test_future = Future("TestData")

        def test_method():
            pass

        test_future.freeze(test_method)

        assert test_future.method == test_method
        assert test_future.args == ()
        assert test_future.kwdargs == {}

    def test_thaw_suppress_exception_no_exception(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            return True

        test_future.freeze(test_method)

        expected = True
        actual = test_future.thaw()
        assert expected == actual

        assert test_future.res is True
        assert test_future.evt.is_set() is True

    def test_thaw_suppress_exception_exception(self):
        test_future = Future("TestData")

        def test_method():
            return True

        test_future.freeze(
            test_method, "arg1", "arg2", kwarg1="test", kwarg2="test")

        test_result = test_future.thaw()

        assert isinstance(test_result, TypeError)

        assert isinstance(test_future.res, TypeError)
        assert test_future.evt.is_set() is True

    def test_thaw_not_suppress_exception_no_exception(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            return True

        test_future.freeze(test_method)

        expected = True
        actual = test_future.thaw(False)
        assert expected == actual

        assert test_future.res is True
        assert test_future.evt.is_set() is True

    def test_thaw_not_suppress_exception_raise_exception(self):
        test_future = Future("TestData")

        def test_method():
            return True

        test_future.freeze(
            test_method, "arg1", "arg2", kwarg1="test", kwarg2="test")

        with pytest.raises(TypeError):
            test_future.thaw(False)

        assert test_future.res is None
        assert test_future.evt.is_set() is False

    def test_has_value_unset(self):
        test_future = Future("TestData")

        expected = False
        actual = test_future.has_value()
        assert expected == actual

    def test_has_value_set(self):
        test_future = Future("TestData")

        test_future.evt.set()

        expected = True
        actual = test_future.has_value()
        assert expected == actual

    def test_resolve(self):
        test_future = Future("TestData")

        test_future.resolve(True)

        assert test_future.res is True
        assert test_future.evt.is_set() is True

    def test_resolve_callback(self):
        test_future = Future("TestData")

        def test_callback(obj):
            try:
                obj.res = not obj.res
            except Exception:
                pass

        test_future.add_callback('resolved', test_callback)

        test_future.resolve(True)

        # Callback reverses the boolean 'res' value
        assert test_future.res is False
        assert test_future.evt.is_set() is True

    def test_wait(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            time.sleep(2)
            return True

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        expected = True
        actual = test_future.wait()
        assert expected == actual

    def test_wait_timeout(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            time.sleep(2)
            return True

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        with pytest.raises(TimeoutError):
            test_future.wait(1)

    def test_get_value_block_no_timeout(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            time.sleep(2)
            return True

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        expected = True
        actual = test_future.get_value()
        assert expected == actual

    def test_get_value_block_timeout(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            time.sleep(2)
            return True

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        with pytest.raises(TimeoutError):
            test_future.get_value(True, 1)

    def test_get_value_no_block_fail(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            time.sleep(2)
            return True

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        with pytest.raises(TimeoutError):
            test_future.get_value(False)

    def test_get_value_no_block_pass(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            return True

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        # Making the test running thread to sleep for a while
        # for the self.future_thread to complete and ensure success
        time.sleep(0.5)

        expected = True
        actual = test_future.get_value()
        assert expected == actual

    def test_get_value_suppress_exception(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            raise _TestError("Test Error")

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        future_value = test_future.get_value(True, None, True)
        assert isinstance(future_value, Exception)

    def test_get_value_no_suppress_exception(self):
        test_future = Future("TestData")

        def test_method(*args, **kwargs):
            raise _TestError("Test Error")

        test_future.freeze(test_method)

        self.future_thread = threading.Thread(target=test_future.thaw)
        self.future_thread.start()

        with pytest.raises(_TestError):
            test_future.get_value()

    def teardown_class(self):
        if self.future_thread is not None:
            self.future_thread.join()

# END



#! /usr/bin/env python
#
# ipg.py -- Module for simple FITS viewer in an HTML5 canvas web browser.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
This example illustrates using a Ginga widget in a web browser,  All the
rendering is done on the server side and the browser only acts as a display
front end.  Using this you could create an analysis type environment on a
server and view it via a browser.

See example usage with an ipython notebook at:

    https://gist.github.com/ejeschke/6067409

You will need a reasonably modern web browser with HTML5 canvas support.
Tested with Chromium 41.0.2272.76, Firefox 37.0.2, Safari 7.1.6
"""

import sys
import os
import logging
import threading

import tornado.web
import tornado.ioloop

from ginga import AstroImage
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.misc import log, Task, Bunch
from ginga.Bindings import ImageViewBindings
from ginga.misc.Settings import SettingGroup
from ginga.util.paths import ginga_home
from ginga.util import loader

from ginga.web.pgw import js, PgHelp, Widgets, Viewers


class BasicCanvasView(Viewers.CanvasView):

    def build_gui(self, container):
        """
        This is responsible for building the viewer's UI.  It should
        place the UI in `container`.  Override this to make a custom
        UI.
        """
        vbox = Widgets.VBox()
        vbox.set_border_width(0)

        w = Viewers.GingaViewerWidget(viewer=self)
        vbox.add_widget(w, stretch=1)

        # need to put this in an hbox with an expanding label or the
        # browser wants to resize the canvas, distorting it
        hbox = Widgets.HBox()
        hbox.add_widget(vbox, stretch=0)
        hbox.add_widget(Widgets.Label(''), stretch=1)

        container.set_widget(hbox)

    def embed(self, width=600, height=650):
        """
        Embed a viewer into a Jupyter notebook.
        """
        from IPython.display import IFrame
        return IFrame(self.url, width, height)

    def open(self, new=1):
        """
        Open this viewer in a new browser window or tab.
        (requires `webbrowser` module)
        """
        import webbrowser
        webbrowser.open(self.url, new=new)

    def show(self, fmt=None):
        """
        Capture the window of a viewer.
        """
        # force any delayed redraws
        # TODO: this really needs to be addressed in get_rgb_image_as_bytes()
        # of the various superclasses, as it affects other backends as well
        self.redraw_now()

        from IPython.display import Image

        if fmt is None:
            # what format are we using for the HTML5 canvas--use that
            settings = self.get_settings()
            fmt = settings.get('html5_canvas_format', 'png')

        return Image(data=bytes(self.get_rgb_image_as_bytes(format=fmt)),
                     format=fmt, embed=True)

    def load_fits(self, filepath):
        """
        Load a FITS file into the viewer.
        """
        image = loader.load_data(filepath, logger=self.logger)
        self.set_image(image)

    load = load_fits

    def load_hdu(self, hdu):
        """
        Load an HDU into the viewer.
        """
        image = AstroImage.AstroImage(logger=self.logger)
        image.load_hdu(hdu)

        self.set_image(image)

    def load_data(self, data_np):
        """
        Load raw numpy data into the viewer.
        """
        image = AstroImage.AstroImage(logger=self.logger)
        image.set_data(data_np)

        self.set_image(image)

    def add_canvas(self, tag=None):
        # add a canvas to the view
        my_canvas = self.get_canvas()
        DrawingCanvas = my_canvas.get_draw_class('drawingcanvas')
        canvas = DrawingCanvas()
        # enable drawing on the canvas
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype(None)
        canvas.ui_set_active(True)
        canvas.set_surface(self)
        canvas.register_for_cursor_drawing(self)
        # add the canvas to the view.
        my_canvas.add(canvas, tag=tag)
        canvas.set_draw_mode(None)

        return canvas

    def set_html5_canvas_format(self, fmt):
        """
        Sets the format used for rendering to the HTML5 canvas.
        'png' offers greater clarity, especially for small text, but
        does not have as good of performance as 'jpeg'.
        """
        fmt = fmt.lower()
        if fmt not in ('jpeg', 'png'):
            raise ValueError("Format must be one of {jpeg|png} not '%s'" % (
                fmt))

        settings = self.get_settings()
        settings.set(html5_canvas_format=fmt)

    def get_html5_canvas_format(self):
        settings = self.get_settings()
        return settings.get('html5_canvas_format')


class EnhancedCanvasView(BasicCanvasView):
    """
    Like BasicCanvasView, but includes a readout widget for when the
    cursor is moved over the canvas to display the coordinates.
    """

    def build_gui(self, container):
        """
        This is responsible for building the viewer's UI.  It should
        place the UI in `container`.
        """
        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(1)

        w = Viewers.GingaViewerWidget(viewer=self)
        vbox.add_widget(w, stretch=1)

        # set up to capture cursor movement for reading out coordinates

        # coordinates reported in base 1 or 0?
        self.pixel_base = 1.0

        self.readout = Widgets.Label("")
        vbox.add_widget(self.readout, stretch=0)

        #self.set_callback('none-move', self.motion_cb)
        self.set_callback('cursor-changed', self.motion_cb)

        # need to put this in an hbox with an expanding label or the
        # browser wants to resize the canvas, distorting it
        hbox = Widgets.HBox()
        hbox.add_widget(vbox, stretch=0)
        hbox.add_widget(Widgets.Label(''), stretch=1)

        container.set_widget(hbox)

    def motion_cb(self, viewer, button, data_x, data_y):

        # Get the value under the data coordinates
        try:
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + 0.5), int(data_y + 0.5))

        except Exception:
            value = None

        pb = self.pixel_base
        fits_x, fits_y = data_x + pb, data_y + pb

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.set_text(text)

    def set_readout_text(self, text):
        self.readout.set_text(text)


class ViewerFactory(object):
    """
    This is a factory class that churns out web viewers for a web
    application.

    The most important method of interest is get_viewer().
    """

    def __init__(self, logger, app, thread_pool):
        """
        Parameters
        ----------
        logger : python compatible logger
            a logging-module compatible logger object
        app : ginga pgw web application object
        thread_pool : a ginga thread pool
        """
        self.logger = logger
        self.app = app
        self.thread_pool = thread_pool
        self.dc = get_canvas_types()
        # dict of viewers
        self.viewers = {}

    def get_threadpool(self):
        return self.thread_pool

    def make_viewer(self, window, viewer_class=None,
                    width=512, height=512):
        if viewer_class is None:
            viewer_class = EnhancedCanvasView

        # load binding preferences if available
        cfgfile = os.path.join(ginga_home, "ipg_bindings.cfg")
        bindprefs = SettingGroup(name='bindings', logger=self.logger,
                                 preffile=cfgfile)
        bindprefs.load(onError='silent')

        bd = ImageViewBindings(self.logger, settings=bindprefs)

        fi = viewer_class(self.logger, bindings=bd)
        fi.url = window.url
        # set up some reasonable defaults--user can change these later
        # if desired
        fi.set_autocut_params('zscale')
        fi.enable_autocuts('on')
        fi.enable_autozoom('on')
        fi.set_bg(0.2, 0.2, 0.2)
        fi.ui_set_active(True)
        fi.ipg_parent = self

        # enable most key/mouse operations
        bd = fi.get_bindings()
        bd.enable_all(True)

        # set up a non-private canvas for drawing
        canvas = self.dc.DrawingCanvas()
        canvas.set_surface(fi)
        # add canvas to view
        private_canvas = fi.get_canvas()
        private_canvas.add(canvas)
        canvas.ui_set_active(True)
        fi.set_canvas(canvas)

        fi.set_desired_size(width, height)
        # force allocation of a surface--may be resized later
        fi.configure_surface(width, height)

        # add little mode indicator that shows modal states in
        # the corner
        fi.show_mode_indicator(True)

        # Have the viewer build it's UI into the container
        fi.build_gui(window)

        v_info = Bunch.Bunch(url=window.url, viewer=fi,
                             top=window)
        return v_info

    def get_viewer(self, v_id, viewer_class=None, width=512, height=512,
                   force_new=False):
        """
        Get an existing viewer by viewer id.  If the viewer does not yet
        exist, make a new one.
        """
        if not force_new:
            try:
                return self.viewers[v_id]
            except KeyError:
                pass

        #  create top level window
        window = self.app.make_window("Viewer %s" % v_id, wid=v_id)

        # We get back a record with information about the viewer
        v_info = self.make_viewer(window, viewer_class=viewer_class,
                                  width=width, height=height)

        # Save it under this viewer id
        self.viewers[v_id] = v_info
        return v_info

    def delete_viewer(self, v_id):
        del self.viewers[v_id]

    def delete_all_viewers(self):
        self.viewers = {}


class FileHandler(tornado.web.RequestHandler):
    """
    This is a handler that is started to allow a REST-type web API to
    create and manipulate viewers.

    Currently it only allows the following commands:
        .../viewer?id=v1&cmd=get           Create/access a viewer
        .../viewer?id=v1&cmd=load&path=... Load the viewer
    """

    def initialize(self, name, factory):
        self.name = name
        self.factory = factory
        self.logger = factory.logger
        self.logger.debug("filehandler initialize")

    def get(self):
        self.logger.debug("filehandler get")
        # Collect arguments
        # TODO: width, height?
        cmd = self.get_argument('cmd', 'get')
        v_id = self.get_argument('id', 'v1')

        v_info = self.factory.get_viewer(v_id)

        if cmd == 'get':
            self._do_get(v_info)

        elif cmd == 'load':
            self._do_load(v_info)

    def _do_get(self, v_info):
        # Get window
        window = v_info.top

        # render back to caller
        output = window.render()
        self.write(output)

    def _do_load(self, v_info):
        path = self.get_argument('path', None)
        if path is not None:
            v_info.viewer.load_fits(path)


class WebServer(object):

    def __init__(self, app, thread_pool, factory,
                 host='localhost', port=9909, ev_quit=None,
                 viewer_class=None):

        self.host = host
        self.port = port
        self.app = app
        self.logger = app.logger
        self.thread_pool = thread_pool
        self.factory = factory
        if ev_quit is None:
            ev_quit = threading.Event()
        self.ev_quit = ev_quit
        self.default_viewer_class = viewer_class

        self.server = None
        self.http_server = None

    def start(self, use_thread=True, no_ioloop=False):
        self.thread_pool.startall()

        js_path = os.path.dirname(js.__file__)

        self.server = tornado.web.Application([
            (r"/js/(.*\.js)", tornado.web.StaticFileHandler,
             {"path": js_path}),
            (r"/viewer", FileHandler,
             dict(name='Ginga', factory=self.factory)),
            (r"/app", PgHelp.WindowHandler,
             dict(name='Application', url='/app', app=self.app)),
            (r"/app/socket", PgHelp.ApplicationHandler,
             dict(name='Ginga', app=self.app)),
        ], factory=self.factory, logger=self.logger)

        self.http_server = self.server.listen(self.port, self.host)

        if no_ioloop:
            self.t_ioloop = None
        else:
            self.t_ioloop = tornado.ioloop.IOLoop.instance()
            if use_thread:
                task = Task.FuncTask2(self.t_ioloop.start)
                self.thread_pool.addTask(task)
            else:
                self.t_ioloop.start()

    def stop(self):
        # how to stop tornado server?
        if self.t_ioloop is not None:
            self.t_ioloop.stop()

        self.thread_pool.stopall()
        self.ev_quit.set()
        # stop and dereference the tornado server
        if self.http_server is not None:
            self.http_server.stop()
            self.http_server = None
        self.server = None

    def get_viewer(self, v_id, viewer_class=None, width=512, height=512,
                   force_new=False):

        if viewer_class is None:
            viewer_class = self.default_viewer_class

        v_info = self.factory.get_viewer(v_id, viewer_class=viewer_class,
                                         width=width, height=height,
                                         force_new=force_new)
        return v_info.viewer


def make_server(logger=None, basedir='.', numthreads=5,
                host='localhost', port=9909, viewer_class=None,
                use_opencv=False):

    if logger is None:
        logger = log.get_logger("ipg", null=True)
    ev_quit = threading.Event()

    if use_opencv:
        from ginga import trcalc
        try:
            trcalc.use('opencv')
        except Exception as e:
            logger.warning("Error using opencv: %s" % str(e))

    thread_pool = Task.ThreadPool(numthreads, logger,
                                  ev_quit=ev_quit)

    base_url = "http://%s:%d/app" % (host, port)
    app = Widgets.Application(logger=logger, base_url=base_url,
                              host=host, port=port)

    factory = ViewerFactory(logger, app, thread_pool)

    server = WebServer(app, thread_pool, factory,
                       host=host, port=port, viewer_class=viewer_class)

    return server


def main(options, args):

    logger = log.get_logger("ipg", options=options)

    server = make_server(logger=logger, basedir=options.basedir,
                         numthreads=options.numthreads, host=options.host,
                         port=options.port, use_opencv=options.use_opencv)
    viewer = server.get_viewer('v1')

    logger.info("Starting server with one viewer, connect at %s" % viewer.url)
    try:
        server.start(use_thread=False)

    except KeyboardInterrupt:
        logger.info("Interrupted!")
        server.stop()

    logger.info("Server terminating ...")


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("-d", "--basedir", dest="basedir", metavar="DIR",
                        default=".",
                        help="Directory which is at the base of file open requests")
    argprs.add_argument("--debug", dest="debug", default=False, action="store_true",
                        help="Enter the pdb debugger on main()")
    argprs.add_argument("--host", dest="host", metavar="HOST",
                        default="localhost",
                        help="HOST used to decide which interfaces to listen on")
    argprs.add_argument("--log", dest="logfile", metavar="FILE",
                        help="Write logging output to FILE")
    argprs.add_argument("--loglevel", dest="loglevel", metavar="LEVEL",
                        type=int, default=logging.INFO,
                        help="Set logging level to LEVEL")
    argprs.add_argument("--numthreads", dest="numthreads", type=int,
                        default=5, metavar="NUM",
                        help="Start NUM threads in thread pool")
    argprs.add_argument("--stderr", dest="logstderr", default=False,
                        action="store_true",
                        help="Copy logging also to stderr")
    argprs.add_argument("--opencv", dest="use_opencv", default=False,
                        action="store_true",
                        help="Use OpenCv acceleration")
    argprs.add_argument("-p", "--port", dest="port",
                        type=int, default=9909, metavar="PORT",
                        help="Default PORT to use for the web socket")
    argprs.add_argument("--profile", dest="profile", action="store_true",
                        default=False,
                        help="Run the profiler on main()")

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    # Are we debugging this?
    if options.debug:
        import pdb

        pdb.run('main(options, args)')

    # Are we profiling this?
    elif options.profile:
        import profile

        print(("%s profile:" % sys.argv[0]))
        profile.run('main(options, args)')

    else:
        main(options, args)

# END



import matplotlib.pyplot as plt

from ginga.mplw.ImageViewCanvasMpl import ImageViewCanvas
from ginga.misc import log
from ginga.AstroImage import AstroImage
from ginga import cmap
# add matplotlib colormaps to ginga's own set
cmap.add_matplotlib_cmaps()

## from IPython.display import Image
## from io import BytesIO


class CustomMplViewer(ImageViewCanvas):

    def get_nb_image(self):
        return AstroImage(
            data=bytes(self.get_rgb_image_as_bytes(format='png')),
            format='png', embed=True)

    def load(self, filepath):
        image = AstroImage(logger=self.logger)
        image.load_file(filepath)

        self.set_image(image)

    def show(self):
        self.figure.show()

    def add_canvas(self, tag=None):
        # add a canvas to the view
        DrawingCanvas = self.getDrawClass('drawingcanvas')
        canvas = DrawingCanvas()
        # enable drawing on the canvas
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.ui_set_active(True)
        canvas.set_surface(self)
        # add the canvas to the view.
        self.add(canvas, tag=tag)
        return canvas


def get_viewer():
    # Set to True to get diagnostic logging output
    use_logger = False
    logger = log.get_logger(null=not use_logger, log_stderr=True)

    # create a regular matplotlib figure
    fig = plt.figure()

    # create a ginga object, initialize some defaults and
    # tell it about the figure
    viewer = CustomMplViewer(logger)
    viewer.enable_autocuts('on')
    viewer.set_autocut_params('zscale')
    viewer.set_figure(fig)

    # enable all interactive ginga features
    viewer.get_bindings().enable_all(True)

    return viewer



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import sys

from ginga.util import iohelper

IS_WINDOWS = sys.platform.startswith('win')


def test_get_fileinfo_real_file():
    """Test behavior on real file."""
    bnch = iohelper.get_fileinfo(iohelper.__file__)
    assert bnch['ondisk']
    assert bnch['name'] == 'iohelper'
    assert bnch['numhdu'] is None
    # filepath and url are OS-dependent, hence the lax check here
    assert bnch['url'].startswith('file://')


def test_get_fileinfo_dummy_file():
    """Test behavior on dummy file."""

    # pathlib behaves differently depending on OS.
    if IS_WINDOWS:
        filename = r'C:\mypath\dummyfile.fits[1]'
        filepath = 'C:\\mypath\\dummyfile.fits'
        url = 'file:///C:/mypath/dummyfile.fits'
    else:
        filename = '/mypath/dummyfile.fits[1]'
        filepath = '/mypath/dummyfile.fits'
        url = 'file:///mypath/dummyfile.fits'

    bnch = iohelper.get_fileinfo(filename)
    assert not bnch['ondisk']
    assert bnch['name'] == 'dummyfile[1]'
    assert bnch['numhdu'] == 1
    assert bnch['filepath'] == filepath
    assert bnch['url'] == url



#
# PlotView.py -- base class for plot viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import logging
import numpy as np

from ginga.misc import Callback, Settings
from ginga import AstroImage

from ginga.gw import Widgets
try:
    from ginga.gw import Plot
    from ginga.util import plots
    have_mpl = True
except ImportError:
    have_mpl = False


class PlotViewGw(Callback.Callbacks):
    """A Ginga viewer for displaying 2D plots using matplotlib.
    """

    vname = 'Ginga Plot'
    vtypes = [AstroImage.AstroImage]

    @classmethod
    def viewable(cls, dataobj):
        """Test whether `dataobj` is viewable by this viewer."""
        if not isinstance(dataobj, AstroImage.AstroImage):
            return False
        shp = list(dataobj.shape)
        if 0 in shp or len(shp) != 1:
            return False
        return True

    def __init__(self, logger=None, settings=None):
        Callback.Callbacks.__init__(self)

        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.Logger('PlotView')

        # Create settings and set defaults
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings
        self.settings.add_defaults(plot_bg='white', show_marker=False,
                                   linewidth=1, linestyle='-',
                                   linecolor='blue', markersize=6,
                                   markerwidth=0.5, markercolor='red',
                                   markerstyle='o', file_suffix='.png')

        # for debugging
        self.name = str(self)

        if not have_mpl:
            raise ImportError('Install matplotlib to use this plugin')

        top = Widgets.VBox()
        top.set_border_width(4)

        self.line_plot = plots.Plot(logger=self.logger,
                                    width=400, height=400)
        bg = self.settings.get('plot_bg', 'white')
        if plots.MPL_GE_2_0:
            kwargs = {'facecolor': bg}
        else:
            kwargs = {'axisbg': bg}
        self.line_plot.add_axis(**kwargs)
        self.plot_w = Plot.PlotWidget(self.line_plot)
        self.plot_w.resize(400, 400)

        # enable interactivity in the plot
        self.line_plot.connect_ui()
        self.line_plot.enable(zoom=True, pan=True)
        self.line_plot.add_callback('limits-set', self.limits_cb)

        ax = self.line_plot.ax
        ax.grid(True)

        top.add_widget(self.plot_w, stretch=1)

        captions = (('Log X', 'checkbutton', 'Log Y', 'checkbutton',
                     'Show Marker', 'checkbutton'),
                    ('X Low:', 'label', 'x_lo', 'entry',
                     'X High:', 'label', 'x_hi', 'entry',
                     'Reset X', 'button'),
                    ('Y Low:', 'label', 'y_lo', 'entry',
                     'Y High:', 'label', 'y_hi', 'entry',
                     'Reset Y', 'button'),
                    ('Save', 'button'))
        # for now...
        orientation = 'vertical'
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w = b

        top.add_widget(w, stretch=0)

        b.log_x.set_state(self.line_plot.logx)
        b.log_x.add_callback('activated', self.log_x_cb)
        b.log_x.set_tooltip('Plot X-axis in log scale')

        b.log_y.set_state(self.line_plot.logy)
        b.log_y.add_callback('activated', self.log_y_cb)
        b.log_y.set_tooltip('Plot Y-axis in log scale')

        b.x_lo.add_callback('activated', lambda w: self.set_xlim_cb())
        b.x_lo.set_tooltip('Set X lower limit')

        b.x_hi.add_callback('activated', lambda w: self.set_xlim_cb())
        b.x_hi.set_tooltip('Set X upper limit')

        b.y_lo.add_callback('activated', lambda w: self.set_ylim_cb())
        b.y_lo.set_tooltip('Set Y lower limit')

        b.y_hi.add_callback('activated', lambda w: self.set_ylim_cb())
        b.y_hi.set_tooltip('Set Y upper limit')

        b.reset_x.add_callback('activated', lambda w: self.reset_xlim_cb())
        b.reset_x.set_tooltip('Autoscale X limits')

        b.reset_y.add_callback('activated', lambda w: self.reset_ylim_cb())
        b.reset_y.set_tooltip('Autoscale Y limits')

        b.show_marker.set_state(self.settings.get('show_marker', False))
        b.show_marker.add_callback('activated', self.set_marker_cb)
        b.show_marker.set_tooltip('Mark data points')

        # Button to save plot
        self.save_plot = b.save
        self.save_plot.set_tooltip('Save table plot')
        self.save_plot.add_callback('activated', lambda w: self.save_cb())
        self.save_plot.set_enabled(False)

        self.widget = top

        # For callbacks
        for name in ['image-set']:
            self.enable_callback(name)

    def get_widget(self):
        return self.widget

    def get_settings(self):
        return self.settings

    def get_logger(self):
        return self.logger

    def clear(self):
        self.widget.clear()

    def initialize_channel(self, fv, channel):
        # no housekeeping to do (for now) on our part, just override to
        # suppress the logger warning
        pass

    def set_dataobj(self, dataobj):
        if not self.viewable(dataobj):
            raise ValueError("Can't display this data object")

        self._dataobj = dataobj

        self.do_plot(reset_xlimits=True, reset_ylimits=True)

        self.make_callback('image-set', dataobj)

    def get_dataobj(self):
        return self._dataobj

    def clear_data(self):
        """Clear comboboxes and columns."""
        self.w.x_lo.set_text('')
        self.w.x_hi.set_text('')
        self.w.y_lo.set_text('')
        self.w.y_hi.set_text('')

    def clear_plot(self):
        """Clear plot display."""
        self.line_plot.clear()
        self.line_plot.draw()
        self.save_plot.set_enabled(False)

    def do_plot(self, reset_xlimits=True, reset_ylimits=True):
        """Simple line plot."""
        self.clear_plot()

        if self._dataobj is None:  # No data to plot
            return

        plt_kw = {
            'lw': self.settings.get('linewidth', 1),
            'ls': self.settings.get('linestyle', '-'),
            'color': self.settings.get('linecolor', 'blue'),
            'ms': self.settings.get('markersize', 6),
            'mew': self.settings.get('markerwidth', 0.5),
            'mfc': self.settings.get('markercolor', 'red')}
        plt_kw['mec'] = plt_kw['mfc']

        try:
            x_data, y_data = self.get_plot_data()
            marker = self.get_marker()

            self.line_plot.plot(
                x_data, y_data,
                xtitle=self.get_label('x'), ytitle=self.get_label('y'),
                marker=marker, **plt_kw)

            if not reset_xlimits:
                self.set_xlim_cb()
            self.set_xlimits_widgets()

            if not reset_ylimits:
                self.set_ylim_cb()
            self.set_ylimits_widgets()

        except Exception as e:
            self.logger.error(str(e))
        else:
            self.save_plot.set_enabled(True)

    def set_xlimits_widgets(self, set_min=True, set_max=True):
        """Populate axis limits GUI with current plot values."""
        xmin, xmax = self.line_plot.ax.get_xlim()
        if set_min:
            self.w.x_lo.set_text('{0}'.format(xmin))
        if set_max:
            self.w.x_hi.set_text('{0}'.format(xmax))

    def set_ylimits_widgets(self, set_min=True, set_max=True):
        """Populate axis limits GUI with current plot values."""
        ymin, ymax = self.line_plot.ax.get_ylim()
        if set_min:
            self.w.y_lo.set_text('{0}'.format(ymin))
        if set_max:
            self.w.y_hi.set_text('{0}'.format(ymax))

    def limits_cb(self, plot, dct):
        """Callback that is called when the limits are set by the
        plot object.
        """
        self.set_xlimits_widgets()
        self.set_ylimits_widgets()

    def get_plot_data(self):
        """Extract only good data point for plotting."""
        y_data = self._dataobj.get_data()
        x_data = np.arange(len(y_data))

        return x_data, y_data

    def get_marker(self):
        _marker_type = self.settings.get('markerstyle', 'o')

        if not self.w.show_marker.get_state():
            _marker_type = None

        return _marker_type

    def get_label(self, axis):
        """Return plot label for the given axis."""

        if axis == 'x':
            label = 'Index'
        if axis == 'y':
            label = 'Value'

        return label

    def log_x_cb(self, w, val):
        """Toggle linear/log scale for X-axis."""
        self.line_plot.logx = val
        self.do_plot()

    def log_y_cb(self, w, val):
        """Toggle linear/log scale for Y-axis."""
        self.line_plot.logy = val
        self.do_plot()

    def set_xlim_cb(self, redraw=True):
        """Set plot limit based on user values."""
        try:
            xmin = float(self.w.x_lo.get_text())
        except Exception:
            set_min = True
        else:
            set_min = False

        try:
            xmax = float(self.w.x_hi.get_text())
        except Exception:
            set_max = True
        else:
            set_max = False

        if set_min or set_max:
            self.line_plot.draw()
            self.set_xlimits_widgets(set_min=set_min, set_max=set_max)

        if not (set_min and set_max):
            self.line_plot.ax.set_xlim(xmin, xmax)
            if redraw:
                self.line_plot.draw()

    def set_ylim_cb(self, redraw=True):
        """Set plot limit based on user values."""
        try:
            ymin = float(self.w.y_lo.get_text())
        except Exception:
            set_min = True
        else:
            set_min = False

        try:
            ymax = float(self.w.y_hi.get_text())
        except Exception:
            set_max = True
        else:
            set_max = False

        if set_min or set_max:
            self.line_plot.draw()
            self.set_ylimits_widgets(set_min=set_min, set_max=set_max)

        if not (set_min and set_max):
            self.line_plot.ax.set_ylim(ymin, ymax)
            if redraw:
                self.line_plot.draw()

    def reset_xlim_cb(self):
        self.line_plot.autoscale('x')

    def reset_ylim_cb(self):
        self.line_plot.autoscale('y')

    def set_marker_cb(self, w, val):
        """Toggle show/hide data point markers."""
        self.do_plot()

    def save_cb(self):
        """Save plot to file."""

        # This just defines the basename.
        # Extension has to be explicitly defined or things can get messy.
        w = Widgets.SaveDialog(title='Save plot')
        target = w.get_path()
        if target is None:
            # Save canceled
            return

        plot_ext = self.settings.get('file_suffix', '.png')

        if not target.endswith(plot_ext):
            target += plot_ext

        # TODO: This can be a user preference?
        fig_dpi = 100

        try:
            fig = self.line_plot.get_figure()
            fig.savefig(target, dpi=fig_dpi)

        except Exception as e:
            self.logger.error(str(e))
        else:
            self.logger.info('Table plot saved as {0}'.format(target))

    def __str__(self):
        return "PlotViewer"



from ginga import toolkit
tkname = toolkit.get_family()

if tkname == 'qt':
    from ginga.qtw.Viewers import *  # noqa

elif tkname == 'gtk3':
    from ginga.gtk3w.Viewers import *  # noqa

elif tkname == 'pg':
    from ginga.web.pgw.Viewers import *  # noqa

from ginga.table.TableView import TableViewGw  # noqa

from .PlotView import PlotViewGw  # noqa
